#!/usr/bin/env python3
"""Create a commit, including the CoLRev report."""
from __future__ import annotations

import importlib
import os
import sys
import typing
from importlib.metadata import version
from pathlib import Path

import git
import gitdb.exc

import colrev.env.utils
import colrev.exceptions as colrev_exceptions

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.review_manager
    import colrev.ops.status


class Commit:
    """Create commits"""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments

    ext_script_name: str
    ext_script_version: str
    python_version: str
    git_version: str
    docker_version: str
    last_commit_sha: str
    tree_hash: str
    records_committed: bool
    completeness_condition: bool
    # Note: last_commit_sha and tree_hash are used in the commit message (external template)

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        msg: str,
        manual_author: bool,
        script_name: str,
        saved_args: typing.Optional[dict] = None,
        skip_hooks: bool = False,
    ) -> None:
        self.review_manager = review_manager
        self.msg = msg
        self.manual_author = manual_author
        self.script_name = self._parse_script_name(script_name)
        self.saved_args = self._parse_saved_args(saved_args)
        self.skip_hooks = skip_hooks

        self._set_versions()
        self._set_script_information(script_name)

    def _parse_script_name(self, script_name: str) -> str:
        if script_name == "MANUAL":
            script_name = "Commit created manually or by external script"
        elif " " in script_name:
            script_name = script_name.replace("colrev cli", "colrev").replace(
                "prescreen_cli", "prescreen"
            )
        return script_name

    def _parse_saved_args(self, saved_args: typing.Optional[dict] = None) -> str:
        saved_args_str = ""
        if saved_args is not None:
            for key, value in saved_args.items():
                if isinstance(value, (bool, float, int, str)):
                    if value == "":  # pragma: no cover
                        saved_args_str += f"     --{key} \\\n"
                    else:
                        saved_args_str += f"     --{key}={value} \\\n"
            # Replace the last backslash (for argument chaining across linebreaks)
            saved_args_str = saved_args_str.rstrip(" \\\n")
        return saved_args_str

    def _set_versions(self) -> None:
        self.colrev_version = f'version {version("colrev")}'
        sys_v = sys.version
        self.python_version = f'version {sys_v[: sys_v.find(" ")]}'
        stream = os.popen("git --version")
        self.git_version = stream.read().replace("git ", "").replace("\n", "")
        stream = os.popen("docker --version")
        self.docker_version = stream.read().replace("Docker ", "").replace("\n", "")
        if self.docker_version == "":  # pragma: no cover
            self.docker_version = "Not installed"

    def _set_script_information(self, script_name: str) -> None:
        self.ext_script_name = ""
        self.ext_script_version = ""

        if script_name == "":
            return

        # Note : the script_name / ext_script seems to be different on macos during init!?
        ext_script = script_name.split(" ")[0]
        if ext_script != "colrev":
            try:
                script_version = version(ext_script)
                self.ext_script_name = script_name
                self.ext_script_version = f"version {script_version}"
            except importlib.metadata.PackageNotFoundError:
                pass
            except ValueError:  # pragma: no cover
                # Note : macos error in importlib.metadata.version
                # ValueError: A distribution name is required
                self.ext_script_name = "unknown"
                self.ext_script_version = "unknown"

    def _get_version_flag(self) -> str:
        flag = ""
        if "dirty" in version("colrev"):  # pragma: no cover
            flag = "*"
        return flag

    def _get_commit_report(self, status_operation: colrev.ops.status.Status) -> str:
        report = self._get_commit_report_header()
        report += status_operation.get_review_status_report(colors=False)
        report += self._get_commit_report_details()
        return report

    def _get_commit_report_header(self) -> str:
        template = colrev.env.utils.get_template("ops/commit/commit_report_header.txt")
        content = template.render(commit_details=self)
        return content

    def _get_commit_report_details(self) -> str:
        template = colrev.env.utils.get_template("ops/commit/commit_report_details.txt")
        content = template.render(commit_details=self)
        return content

    def _get_detailed_processing_report(self) -> str:
        processing_report = ""
        report_path = self.review_manager.paths.report
        if report_path.is_file():
            processing_report = "\nProcessing report\n"
            processing_report += "".join(report_path.read_text())
        return processing_report

    def create(self, *, skip_status_yaml: bool = False) -> bool:
        """Create a commit (including the commit message and details)"""
        status_operation = self.review_manager.get_status_operation()
        git_repo = self.review_manager.dataset.get_repo()
        try:
            if not git_repo.index.diff("HEAD"):
                self.review_manager.logger.debug(
                    "No staged changes / cannot create commit"
                )
                return False
        except gitdb.exc.BadName:
            self.review_manager.logger.debug(
                "HEAD does not exist, cannot create commit"
            )

        self.review_manager.logger.debug("Prepare commit: checks and updates")
        if not skip_status_yaml:
            status_yml = self.review_manager.paths.status
            self.review_manager.update_status_yaml()
            self.review_manager.dataset.add_changes(status_yml)

        committer, email = self.review_manager.get_committer()

        if self.manual_author:
            git_author = git.Actor(committer, email)
        else:
            git_author = git.Actor(f"script:{self.script_name}", email)

        # Note : this should run as the last command before creating the commit
        # to ensure that the git tree_hash is up-to-date.
        self.tree_hash = self.review_manager.dataset.get_tree_hash()
        try:
            self.last_commit_sha = self.review_manager.dataset.get_last_commit_sha()
        except ValueError:
            pass

        self.records_committed = self.review_manager.paths.records.is_file()
        self.completeness_condition = self.review_manager.get_completeness_condition()

        self.msg = (
            self.msg
            + self._get_version_flag()
            + self._get_commit_report(status_operation)
            + self._get_detailed_processing_report()
        )
        git_repo.index.commit(
            self.msg,
            author=git_author,
            committer=git.Actor(committer, email),
            skip_hooks=self.skip_hooks,
        )

        self.review_manager.logger.info("Created commit")
        self.review_manager.reset_report_logger()

        if self.review_manager.dataset.has_record_changes():
            if not self.review_manager.force_mode:
                raise colrev_exceptions.DirtyRepoAfterProcessingError(
                    "A clean repository is expected."
                )
            self.review_manager.logger.warning("No clean repository after commit.")

        return True

    def update_report(self, *, msg_file: Path) -> None:
        """Update the report"""
        status_operation = self.review_manager.get_status_operation()
        report = self._get_commit_report(status_operation)
        with open(msg_file, "a", encoding="utf8") as file:
            file.write(report)



#! /usr/bin/env python3
"""CoLRev status stats."""
from __future__ import annotations

import typing

from pydantic import BaseModel

import colrev.process.operation
from colrev.constants import Fields
from colrev.constants import RecordState
from colrev.process.model import ProcessModel


class StatusStatsCurrently(BaseModel):
    """The current status statistics"""

    # pylint: disable=too-many-instance-attributes
    md_retrieved: int
    md_imported: int
    md_prepared: int
    md_processed: int
    rev_prescreen_excluded: int
    rev_prescreen_included: int
    pdf_not_available: int
    pdf_imported: int
    pdf_prepared: int
    rev_excluded: int
    rev_included: int
    rev_synthesized: int
    pdf_needs_retrieval: int
    non_completed: int
    exclusion: dict
    md_needs_manual_preparation: int
    pdf_needs_manual_retrieval: int
    pdf_needs_manual_preparation: int


class StatusStatsOverall(BaseModel):
    """The overall-status statistics (records currently/previously in each state)"""

    # pylint: disable=too-many-instance-attributes
    md_retrieved: int
    md_imported: int
    md_prepared: int
    md_processed: int
    rev_prescreen_excluded: int
    rev_prescreen_included: int
    rev_prescreen: int
    pdf_not_available: int
    pdf_imported: int
    pdf_prepared: int
    rev_screen: int
    rev_excluded: int
    rev_included: int
    rev_synthesized: int


def _get_origin_states_dict(records: dict) -> dict:
    current_origin_states_dict = {}
    for record_dict in records.values():
        for origin in record_dict[Fields.ORIGIN]:
            current_origin_states_dict[origin] = record_dict[Fields.STATUS]
    return current_origin_states_dict


def _get_screening_statistics(
    *,
    review_manager: colrev.review_manager.ReviewManager,
    records: dict,
) -> dict:
    screening_criteria = [
        x[Fields.SCREENING_CRITERIA]
        for x in records.values()
        if x.get(Fields.SCREENING_CRITERIA, "") not in ["", "NA"]
    ]
    criteria = list(review_manager.settings.screen.criteria.keys())
    screening_statistics = {crit: 0 for crit in criteria}
    for screening_case in screening_criteria:
        for criterion in screening_case.split(";"):
            criterion_name, decision = criterion.split("=")
            if decision == "out":
                screening_statistics[criterion_name] += 1
    return screening_statistics


def _get_duplicates_removed(records: dict) -> int:
    md_duplicates_removed = 0
    for record_dict in records.values():
        md_duplicates_removed += (
            len([o for o in record_dict[Fields.ORIGIN] if not o.startswith("md_")]) - 1
        )
    return md_duplicates_removed


def _get_nr_origins(records: dict) -> int:
    origin_list = [x[Fields.ORIGIN] for x in records.values()]
    nr_origins = 0
    for origin in origin_list:
        nr_origins += len([o for o in origin if not o.startswith("md_")])
    return nr_origins


def _get_nr_incomplete(origin_states_dict: dict) -> int:
    """Get the number of incomplete records"""
    return len(
        [
            x
            for x in list(origin_states_dict.values())
            if x
            not in [
                RecordState.rev_synthesized,
                RecordState.rev_excluded,
                RecordState.rev_prescreen_excluded,
                RecordState.pdf_not_available,
            ]
        ]
    )


def _get_freq(*, status_list: list, colrev_status: RecordState) -> int:
    return len([x for x in status_list if colrev_status == x])


def _get_status_stats_currently(
    status_list: list, records: dict, screening_statistics: dict, md_retrieved: int
) -> StatusStatsCurrently:
    precreen_included = _get_freq(
        status_list=status_list, colrev_status=RecordState.rev_prescreen_included
    )
    data = {
        "md_retrieved": md_retrieved,
        "md_imported": _get_freq(
            status_list=status_list, colrev_status=RecordState.md_imported
        ),
        "md_prepared": _get_freq(
            status_list=status_list, colrev_status=RecordState.md_prepared
        ),
        "md_processed": _get_freq(
            status_list=status_list, colrev_status=RecordState.md_processed
        ),
        "rev_prescreen_excluded": _get_freq(
            status_list=status_list, colrev_status=RecordState.rev_prescreen_excluded
        ),
        "rev_prescreen_included": precreen_included,
        "pdf_imported": _get_freq(
            status_list=status_list, colrev_status=RecordState.pdf_imported
        ),
        "pdf_not_available": _get_freq(
            status_list=status_list, colrev_status=RecordState.pdf_not_available
        ),
        "pdf_prepared": _get_freq(
            status_list=status_list, colrev_status=RecordState.pdf_prepared
        ),
        "rev_excluded": _get_freq(
            status_list=status_list, colrev_status=RecordState.rev_excluded
        ),
        "rev_included": _get_freq(
            status_list=status_list, colrev_status=RecordState.rev_included
        ),
        "rev_synthesized": _get_freq(
            status_list=status_list, colrev_status=RecordState.rev_synthesized
        ),
        "pdf_needs_retrieval": precreen_included,
        "exclusion": screening_statistics,
        "md_needs_manual_preparation": _get_freq(
            status_list=status_list,
            colrev_status=RecordState.md_needs_manual_preparation,
        ),
        "pdf_needs_manual_retrieval": _get_freq(
            status_list=status_list,
            colrev_status=RecordState.pdf_needs_manual_retrieval,
        ),
        "pdf_needs_manual_preparation": _get_freq(
            status_list=status_list,
            colrev_status=RecordState.pdf_needs_manual_preparation,
        ),
        "non_completed": len(
            [
                r
                for r in records.values()
                if r[Fields.STATUS]
                not in [
                    RecordState.rev_synthesized,
                    RecordState.rev_prescreen_excluded,
                    RecordState.pdf_not_available,
                    RecordState.rev_excluded,
                ]
            ]
        ),
    }

    return StatusStatsCurrently(**data)


def _get_cumulative_freq(*, status_list: list, colrev_status: RecordState) -> int:
    return len(
        [
            x
            for x in status_list
            if x in RecordState.get_post_x_states(state=colrev_status)
        ]
    )


def _get_md_retrieved(sources: list) -> int:
    md_retrieved = 0
    for source in sources:
        if not source.is_md_source():
            nr_in_file = colrev.loader.load_utils.get_nr_records(source.filename)
            md_retrieved += nr_in_file
    return md_retrieved


def _get_currently_md_retrieved(origin_states_dict: dict, md_retrieved: int) -> int:

    # select records that are not md_ records
    # (origin_states_dict only has records beyond md_retrieved)
    non_md_records = {
        k: v for k, v in origin_states_dict.items() if not k.startswith("md_")
    }

    return md_retrieved - len(non_md_records)


def _get_status_stats_overall(
    status_list: list, records: dict, md_retrieved: int
) -> StatusStatsOverall:

    data = {
        "md_retrieved": md_retrieved,
        "md_imported": len(records.values()),
        "md_prepared": _get_cumulative_freq(
            status_list=status_list, colrev_status=RecordState.md_prepared
        ),
        "md_processed": _get_cumulative_freq(
            status_list=status_list, colrev_status=RecordState.md_processed
        ),
        "rev_prescreen": _get_cumulative_freq(
            status_list=status_list, colrev_status=RecordState.md_processed
        ),
        "rev_prescreen_included": _get_cumulative_freq(
            status_list=status_list, colrev_status=RecordState.rev_prescreen_included
        ),
        "pdf_imported": _get_cumulative_freq(
            status_list=status_list, colrev_status=RecordState.pdf_imported
        ),
        "pdf_prepared": _get_cumulative_freq(
            status_list=status_list, colrev_status=RecordState.pdf_prepared
        ),
        "rev_screen": _get_cumulative_freq(
            status_list=status_list, colrev_status=RecordState.pdf_prepared
        ),
        "rev_included": _get_cumulative_freq(
            status_list=status_list, colrev_status=RecordState.rev_included
        ),
        "rev_synthesized": _get_cumulative_freq(
            status_list=status_list, colrev_status=RecordState.rev_synthesized
        ),
        # Note: temporary states (_man_*) should not be covered in StatusStatsOverall
        "rev_prescreen_excluded": _get_freq(
            status_list=status_list, colrev_status=RecordState.rev_prescreen_excluded
        ),
        "rev_excluded": _get_freq(
            status_list=status_list, colrev_status=RecordState.rev_excluded
        ),
        "pdf_not_available": _get_freq(
            status_list=status_list, colrev_status=RecordState.pdf_not_available
        ),
    }

    return StatusStatsOverall(**data)


def _get_nr_curated_records(
    records: dict, curated: int, overall: StatusStatsOverall
) -> int:
    nr_curated_records = len(
        [
            r
            for r in records.values()
            if colrev.record.record.Record(r).masterdata_is_curated()
        ]
    )
    if curated:  # pragma: no cover
        nr_curated_records = overall.md_processed
    return nr_curated_records


def _get_perc_curated(
    overall: StatusStatsOverall,
    currently: StatusStatsCurrently,
    nr_curated_records: int,
) -> float:
    perc_curated = 0.0
    denominator = (
        overall.md_processed
        + currently.md_prepared
        + currently.md_needs_manual_preparation
        + currently.md_imported
    )
    if denominator > 0:
        perc_curated = int((nr_curated_records / (denominator)) * 100)
    return perc_curated


REQUIRED_ATOMIC_STEPS = {
    RecordState.md_retrieved: 1,
    RecordState.md_imported: 2,
    # note: md_needs_manual_preparation: prep-operation not yet completed successfully.
    RecordState.md_needs_manual_preparation: 2,
    RecordState.md_prepared: 3,
    RecordState.md_processed: 4,
    RecordState.rev_prescreen_included: 5,
    RecordState.rev_prescreen_excluded: 5,
    RecordState.pdf_needs_manual_retrieval: 5,
    RecordState.pdf_imported: 6,
    RecordState.pdf_needs_manual_preparation: 6,
    RecordState.pdf_not_available: 7,
    RecordState.pdf_prepared: 7,
    RecordState.rev_excluded: 8,
    RecordState.rev_included: 8,
    RecordState.rev_synthesized: 9,
}


def _get_completed_atomic_steps(
    records: dict, currently: StatusStatsCurrently, md_duplicates_removed: int
) -> int:
    """Get the number of completed atomic steps"""
    completed_steps = 0
    for record_dict in records.values():
        completed_steps += REQUIRED_ATOMIC_STEPS[record_dict[Fields.STATUS]]
    completed_steps += 4 * md_duplicates_removed
    completed_steps += currently.md_retrieved  # not in records
    return completed_steps


def _get_atomic_steps(
    overall: StatusStatsOverall,
    currently: StatusStatsCurrently,
    md_duplicates_removed: int,
) -> int:
    return (
        # initially, all records have to pass 9 operations (including search)
        9 * overall.md_retrieved
        # for removed duplicates, 5 operations are no longer needed
        - 5 * md_duplicates_removed
        # for rev_prescreen_excluded, 4 operations are no longer needed
        - 4 * currently.rev_prescreen_excluded
        - 3 * currently.pdf_not_available
        - currently.rev_excluded
    )


# pylint: disable=no-member
def get_status_stats(
    *,
    review_manager: colrev.review_manager.ReviewManager,
    records: dict,
) -> StatusStats:
    """Get the status statistics"""

    origin_states_dict = _get_origin_states_dict(records)

    screening_statistics = _get_screening_statistics(
        review_manager=review_manager, records=records
    )
    status_list = [x[Fields.STATUS] for x in records.values()]
    sources = review_manager.settings.sources
    md_retrieved = _get_md_retrieved(sources)
    currently = _get_status_stats_currently(
        status_list, records, screening_statistics, md_retrieved
    )
    overall = _get_status_stats_overall(status_list, records, md_retrieved)
    currently.md_retrieved = _get_currently_md_retrieved(
        origin_states_dict, md_retrieved
    )
    nr_curated_records = _get_nr_curated_records(
        records, review_manager.settings.is_curated_masterdata_repo(), overall
    )
    md_duplicates_removed = _get_duplicates_removed(records)
    nr_incomplete = _get_nr_incomplete(origin_states_dict)

    data = {
        "screening_statistics": screening_statistics,
        "md_duplicates_removed": md_duplicates_removed,
        "nr_origins": _get_nr_origins(records),
        "nr_incomplete": nr_incomplete,
        "overall": overall,
        "currently": currently,
        "nr_curated_records": nr_curated_records,
        "perc_curated": _get_perc_curated(overall, currently, nr_curated_records),
        "completed_atomic_steps": _get_completed_atomic_steps(
            records, currently, md_duplicates_removed
        ),
        "atomic_steps": _get_atomic_steps(overall, currently, md_duplicates_removed),
        "completeness_condition": (nr_incomplete == 0)
        and (currently.md_retrieved == 0),
        "origin_states_dict": origin_states_dict,
    }

    return StatusStats(**data)


class StatusStats(BaseModel):
    """Data class for status statistics"""

    # pylint: disable=too-many-instance-attributes
    atomic_steps: int
    nr_curated_records: int
    perc_curated: float
    md_duplicates_removed: int
    currently: StatusStatsCurrently
    overall: StatusStatsOverall
    completed_atomic_steps: int
    completeness_condition: bool
    nr_incomplete: int
    nr_origins: int
    screening_statistics: dict

    origin_states_dict: dict

    def get_active_metadata_operation_info(self) -> str:
        """Get active metadata operation info (convenience function for status printing)"""
        infos = []
        if self.currently.md_retrieved > 0:
            infos.append(f"{self.currently.md_retrieved} to load")
        if self.currently.md_imported > 0:
            infos.append(f"{self.currently.md_imported} to prepare")
        if self.currently.md_needs_manual_preparation > 0:
            infos.append(
                f"{self.currently.md_needs_manual_preparation} to prepare manually"
            )
        if self.currently.md_prepared > 0:
            infos.append(f"{self.currently.md_prepared} to deduplicate")
        return ", ".join(infos)

    def get_active_pdf_operation_info(self) -> str:
        """Get active PDF operation info (convenience function for status printing)"""
        infos = []
        if self.currently.rev_prescreen_included > 0:
            infos.append(f"{self.currently.rev_prescreen_included} to retrieve")
        if self.currently.pdf_needs_manual_retrieval > 0:
            infos.append(
                f"{self.currently.pdf_needs_manual_retrieval} to retrieve manually"
            )
        if self.currently.pdf_imported > 0:  # pragma: no cover
            infos.append(f"{self.currently.pdf_imported} to prepare")
        if self.currently.pdf_needs_manual_preparation > 0:  # pragma: no cover
            infos.append(
                f"{self.currently.pdf_needs_manual_preparation} to prepare manually"
            )
        return ", ".join(infos)

    def get_transitioned_records(
        self, review_manager: colrev.review_manager.ReviewManager
    ) -> list[typing.Dict]:
        """Get the transitioned records"""

        committed_origin_states_dict = (
            review_manager.dataset.get_committed_origin_state_dict()
        )

        transitioned_records = []
        for (
            committed_origin,
            committed_colrev_status,
        ) in committed_origin_states_dict.items():
            transitioned_record = {
                "origin": committed_origin,
                "source": committed_colrev_status,
                "dest": self.origin_states_dict.get(
                    committed_origin, "no_source_state"
                ),
            }

            if transitioned_record["source"] == transitioned_record["dest"]:
                continue  # no_transition

            operations_type = [
                x["trigger"]
                for x in ProcessModel.transitions
                if (
                    x["source"] == transitioned_record["source"]
                    and x["dest"] == transitioned_record["dest"]
                )
                # Allow for reverse transitions
                or (
                    x["source"] == transitioned_record["dest"]
                    and x["dest"] == transitioned_record["source"]
                )
            ]
            if len(operations_type) == 0:
                transitioned_record["type"] = "invalid_transition"
            else:
                transitioned_record["type"] = operations_type[0]

            transitioned_records.append(transitioned_record)

        return transitioned_records

    def get_priority_operations(self) -> list:
        """Get the priority operations"""

        # get "earliest" states (going backward)
        earliest_state = []
        search_states = [RecordState.rev_synthesized]
        while True:
            if any(
                search_state in self.origin_states_dict.values()
                for search_state in search_states
            ):
                earliest_state = [
                    search_state
                    for search_state in search_states
                    if search_state in self.origin_states_dict.values()
                ]
            search_states = [
                x["source"]  # type: ignore
                for x in ProcessModel.transitions
                if x["dest"] in search_states
            ]
            if [] == search_states:
                break

        # next: get the priority transition for the earliest states
        priority_transitions = [
            x["trigger"]
            for x in ProcessModel.transitions
            if x["source"] in earliest_state
        ]

        return list(set(priority_transitions))

    def get_active_operations(self) -> list:
        """Get the active processing functions"""

        active_operations: typing.List[str] = []
        for state in set(self.origin_states_dict.values()):
            valid_transitions = ProcessModel.get_valid_transitions(state=state)
            active_operations.extend(valid_transitions)

        return active_operations

    def get_operation_in_progress(self, *, transitioned_records: list) -> set:
        """Get the operation currently in progress"""

        return {str(x["type"]) for x in transitioned_records}



#! /usr/bin/env python
"""SearchSource: SYNERGY-datasets"""
from __future__ import annotations

import datetime
import tempfile
import typing
from pathlib import Path

import pandas as pd
import zope.interface
from git import Repo
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument

# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class SYNERGYDatasetsSearchSource:
    """SYNERGY-datasets

    https://github.com/asreview/synergy-dataset

    csv files available in the releases (output directories) are supported:
    https://github.com/asreview/synergy-dataset/tags

    csv files containing article IDs are supported by source-specific preparation.
    metadata is added based on dois and pubmedid (openalex is not yet supported)

    """

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.synergy_datasets"
    # pylint: disable=colrev-missed-constant-usage
    source_identifier = "ID"
    search_types = [SearchType.API]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.review_manager = source_operation.review_manager
        self.search_source = self.settings_class(**settings)
        self.quality_model = self.review_manager.get_qm()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for SYNERGY-datasets"""

        result = {"confidence": 0.0}

        if "doi,pmid,openalex_id,label_included" in data:
            result["confidence"] = 1.0

        if '"doi","pmid","openalex_id","label_included"' in data:
            result["confidence"] = 1.0

        if (
            "record_id,pubmedID,title,authors,abstract,year,label_included,label_abstract_screening"
            in data
        ):
            result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        params_dict = {}
        if params:
            for item in params.split(";"):
                key, value = item.split("=")
                params_dict[key] = value

        if len(params_dict) == 0:
            operation.review_manager.logger.info("Retrieving available datasets")
            date_now_string = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            temp_path = tempfile.gettempdir() / Path(f"{date_now_string}-synergy")
            temp_path.mkdir()
            Repo.clone_from(
                "https://github.com/asreview/synergy-dataset", temp_path, depth=1
            )
            data_path = temp_path / Path("datasets")
            files = data_path.glob("**/*_ids.csv")
            print("https://github.com/asreview/synergy-dataset")
            print(
                "\n- "
                + "\n- ".join([str(f.parent.name) + "/" + str(f.name) for f in files])
            )
            dataset = input("Enter dataset:")
            params_dict = {"dataset": dataset}

        assert "dataset" in params_dict
        dataset = params_dict["dataset"]
        filename = operation.get_unique_filename(
            file_path_string=f"SYNERGY_{dataset.replace('/', '_').replace('_ids.csv', '')}"
        )
        search_source = colrev.settings.SearchSource(
            endpoint="colrev.synergy_datasets",
            filename=filename,
            search_type=SearchType.API,
            search_parameters={"dataset": dataset},
            comment="",
        )
        operation.add_source_and_search(search_source)
        return search_source

    def _load_dataset(self) -> pd.DataFrame:
        date_now_string = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        temp_path = tempfile.gettempdir() / Path(f"{date_now_string}-synergy")
        temp_path.mkdir()
        Repo.clone_from(
            "https://github.com/asreview/synergy-dataset", temp_path, depth=1
        )
        dataset_name = self.search_source.search_parameters["dataset"]
        dataset_df = pd.read_csv(temp_path / Path("datasets") / dataset_name)

        # check data structure
        if not all(x in dataset_df for x in ["doi", "pmid", "openalex_id"]):
            raise colrev_exceptions.SearchSourceException(
                f"Missing doi/pmid/openalex_id in {dataset_name}"
            )
        if "pmid" not in dataset_df:
            dataset_df["pmid"] = pd.NA
        missing_metadata = (
            dataset_df[["doi", "pmid", "openalex_id"]].isna().all(axis="columns")
        )
        missing_metadata_percentage = missing_metadata.sum() / dataset_df.shape[0]
        if missing_metadata_percentage > 0.1:
            self.review_manager.logger.error(
                f"Missing metadata percentage: {missing_metadata_percentage}"
            )
            input("ENTER to continue anyway")
        else:
            self.review_manager.logger.info(
                f"Missing metadata: {missing_metadata_percentage:.2%}"
            )
        return dataset_df

    def _validate_decisions(self, *, decisions: dict, record: dict) -> None:
        if "doi" in record:
            doi = record["doi"].lower()
            if doi in decisions["doi"]:
                decisions["doi"][doi].append(record["label_included"])
            else:
                decisions["doi"][doi] = [record["label_included"]]

        if "pmid" in record:
            pmid = record["pmid"].lower()
            if pmid in decisions["pmid"]:
                decisions["pmid"][pmid].append(record["label_included"])
            else:
                decisions["pmid"][pmid] = [record["label_included"]]

        if "openalex_id" in record:
            oaid = record["openalex_id"].lower()
            if oaid in decisions["openalex_id"]:
                decisions["openalex_id"][oaid].append(record["label_included"])
            else:
                decisions["openalex_id"][oaid] = [record["label_included"]]

    def _check_quality(self, *, decisions: dict) -> None:
        decisions[Fields.DOI] = {
            d: v
            for d, v in decisions[Fields.DOI].items()
            if len(v) > 1 and len(set(v)) != 1
        }
        decisions["pmid"] = {
            d: v
            for d, v in decisions["pmid"].items()
            if len(v) > 1 and len(set(v)) != 1
        }
        decisions["openalex_id"] = {
            d: v
            for d, v in decisions["openalex_id"].items()
            if len(v) > 1 and len(set(v)) != 1
        }
        if decisions[Fields.DOI] or decisions["pmid"] or decisions["openalex_id"]:
            self.review_manager.logger.error(
                "Errors in dataset: ambiguous inclusion decisions:"
            )
            msg = (
                f"{Colors.RED}"
                + f"- dois: {', '.join(decisions['doi'])}"
                + f"- pmid: {', '.join(decisions['pmid'])}"
                + f"- openalex_id: {', '.join(decisions['openalex_id'])}"
                + f"{Colors.END}"
            )
            if self.review_manager.force_mode:
                print(msg)
            else:
                raise colrev_exceptions.SearchSourceException(msg)

    def _prep_record(self, *, record: dict, ind: int) -> None:
        record[Fields.ID] = ind
        record[Fields.ENTRYTYPE] = "article"
        for k in list(record.keys()):
            if str(record[k]) == "nan":
                del record[k]
        if Fields.DOI in record.keys():
            # pylint: disable=colrev-missed-constant-usage
            record[Fields.DOI] = (
                str(record["doi"]).replace("https://doi.org/", "").upper()
            )
        if "pmid" in record.keys():
            record["pmid"] = str(record["pmid"]).replace(
                "https://pubmed.ncbi.nlm.nih.gov/", ""
            )
        if "openalex_id" in record.keys():
            record["openalex_id"] = str(record["openalex_id"]).replace(
                "https://openalex.org/", ""
            )

    def _validate_source(self) -> None:
        source = self.search_source
        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")
        assert source.search_type == SearchType.API

    def search(self, rerun: bool) -> None:
        """Run a search of the SYNERGY datasets"""

        self._validate_source()

        dataset_df = self._load_dataset()

        synergy_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=False,
        )
        existing_keys = {
            Fields.DOI: [
                r[Fields.DOI]
                for r in synergy_feed.feed_records.values()
                if Fields.DOI in r
            ],
            "pmid": [
                r["pmid"] for r in synergy_feed.feed_records.values() if "pmid" in r
            ],
            "openalex_id": [
                r["openalex_id"]
                for r in synergy_feed.feed_records.values()
                if "openalex_id" in r
            ],
        }

        decisions: typing.Dict[str, typing.Dict[str, list]] = {
            Fields.DOI: {},
            "pmid": {},
            "openalex_id": {},
        }
        empty_records, duplicates = 0, 0
        for ind, record in enumerate(dataset_df.to_dict(orient="records")):
            self._prep_record(record=record, ind=ind)
            self._validate_decisions(decisions=decisions, record=record)
            # Skip records without metadata
            if {Fields.ID, Fields.ENTRYTYPE, "label_included"} == set(record.keys()):
                empty_records += 1
                continue

            # Skip records that are already in the feed
            if any(
                value in existing_keys[key]
                for key, value in record.items()
                if key in [Fields.DOI, "pmid", "openalex_id"]
            ):
                duplicates += 1
                continue

            for key in list(record.keys()):
                if key not in [
                    Fields.ID,
                    Fields.DOI,
                    "pmid",
                    "openalex_id",
                    Fields.ENTRYTYPE,
                ]:
                    record[f"colrev.synergy_datasets.{key}"] = record.pop(key)

            if Fields.DOI in record:
                existing_keys[Fields.DOI].append(record[Fields.DOI])
            if "pmid" in record:
                existing_keys["pmid"].append(record["pmid"])
            if "openalex_id" in record:
                existing_keys["openalex_id"].append(record["openalex_id"])

            synergy_feed.add_update_record(
                retrieved_record=colrev.record.record.Record(record)
            )

            # The linking of doi/... should happen in the prep operation

        self._check_quality(decisions=decisions)
        self.review_manager.logger.info(f"Dropped {empty_records} empty records")
        self.review_manager.logger.info(f"Dropped {duplicates} duplicate records")
        self.review_manager.dataset.load_records_dict()
        synergy_feed.save()

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
            )
            for record in records.values():
                if "pmid" in record:
                    record["pubmedid"] = record["pmid"].replace(
                        "https://pubmed.ncbi.nlm.nih.gov/", ""
                    )
                    del record["pmid"]
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for SYNERGY-datasets"""

        record.rename_field(
            key="colrev.synergy_datasets.pubmedid", new_key=Fields.PUBMED_ID
        )
        record.rename_field(
            key="colrev.synergy_datasets.openalex_id", new_key="colrev.open_alex.id"
        )
        if not any(
            x in record.data
            for x in [Fields.PUBMED_ID, Fields.DOI, "colrev.open_alex.id"]
        ):
            record.prescreen_exclude(reason="no-metadata-available")
        return record



#! /usr/bin/env python
"""SearchSource: GitHub"""
from __future__ import annotations

import re
import typing
from multiprocessing import Lock
from pathlib import Path

import inquirer
import zope.interface
from github import Auth
from github import Github
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.ops.search
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.github.src import record_transformer

# pylint: disable=unused-argument


def is_github_api_key(previous: dict, answer: str) -> bool:
    """Validate GitHub API key format"""
    api_key_pattern = re.compile(r"[a-zA-Z0-9_-]{40}")
    if api_key_pattern.fullmatch(answer):
        return True
    raise inquirer.errors.ValidationError("", reason="Invalid GitHub API key format.")


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class GitHubSearchSource:
    """GitHub API"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    search_types = [SearchType.API]
    endpoint = "colrev.github"
    source_identifier = Fields.URL
    ci_supported: bool = Field(default=True)

    heuristic_status = SearchSourceHeuristicStatus.todo

    docs_link = (
        "https://colrev-environment.github.io/colrev/dev_docs/packages/"
        + "package_interfaces.html#colrev.package_manager.interfaces.SearchSourceInterface"
        + "https://docs.github.com/en/rest?apiVersion=2022-11-28"
    )
    db_url = "https://github.com/"
    SETTINGS = {
        "api_key": "packages.search_source.colrev.github.api_key",
    }
    _github_md_filename = Path("data/search/md_github.bib")

    def __init__(
        self,
        *,
        source_operation: colrev.process.operation.Operation,
        settings: typing.Optional[dict] = None,
    ) -> None:
        self.review_manager = source_operation.review_manager
        if settings:
            # GitHub as a search source
            self.search_source = self.settings_class(**settings)
        else:
            # GitHub as a md-prep source
            github_md_source_l = [
                s
                for s in source_operation.review_manager.settings.sources
                if s.filename == self._github_md_filename
            ]
            if github_md_source_l:
                self.search_source = github_md_source_l[0]
            else:
                self.search_source = colrev.settings.SearchSource(
                    endpoint=self.endpoint,
                    filename=self._github_md_filename,
                    search_type=SearchType.MD,
                    search_parameters={},
                    comment="",
                )
            self.github_lock = Lock()

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for GitHub"""

        result = {"confidence": 0.0}

        return result

    @classmethod
    def _choice_scope(cls) -> str:
        questions = [
            inquirer.Checkbox(
                "search_in",
                message="Where do you want to search?",
                choices=[("URL", "url"), ("Readme", "readme"), ("Topic", "topic")],
            )
        ]

        answers = inquirer.prompt(questions)
        return ",".join(answers["search_in"])

    @classmethod
    def add_endpoint(
        cls, operation: colrev.ops.search.Search, params: str
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        cls._get_api_key()
        params_dict = {}
        if params:
            for item in params.split(";"):
                try:
                    key, value = item.split("=")
                    if key in ["title", "readme"]:
                        params_dict[key] = value
                    else:
                        raise colrev_exceptions.InvalidQueryException(
                            "GitHub search_parameters support title or readme field"
                        )
                except ValueError:
                    print("Invalid search parameter format")

        if len(params_dict) == 0:
            search_source = operation.create_api_source(endpoint="colrev.github")

            # Checking where to search
            search_source.search_parameters["scope"] = cls._choice_scope()

        else:
            if Fields.URL in params_dict:
                query = {
                    "query": (
                        params_dict[Fields.URL]
                        .replace("https://github.com/search?q=%2F", "")
                        .replace("https://github.com/search?q=", "")
                        .replace("&type=repositories", "")
                        .replace("+", " ")
                    )
                }
            else:
                query = params_dict

            filename = operation.get_unique_filename(file_path_string="github")
            search_source = colrev.settings.SearchSource(
                endpoint="colrev.github",
                filename=filename,
                search_type=SearchType.API,
                search_parameters=query,
                comment="",
            )

        operation.add_source_and_search(search_source)
        return search_source

    @classmethod
    def _get_api_key(cls) -> str:
        env_man = colrev.env.environment_manager.EnvironmentManager()
        api_key = env_man.get_settings_by_key(cls.SETTINGS["api_key"])
        if api_key is not None and is_github_api_key({}, api_key):
            return api_key

        questions = [
            inquirer.Text(
                "github_api_key",
                message="Enter your GitHub API key",
                validate=is_github_api_key,
            ),
        ]
        answers = inquirer.prompt(questions)
        input_key = answers["github_api_key"]
        env_man.update_registry(cls.SETTINGS["api_key"], input_key)
        print("API key saved in environment settings.")
        return input_key

    def _run_api_search(
        self, github_feed: colrev.ops.search_api_feed.SearchAPIFeed
    ) -> None:

        if not self.search_source.search_parameters:
            raise ValueError("No search parameters defined for GitHub search source")

        keywords = self.search_source.search_parameters.get("query", "")
        scope = self.search_source.search_parameters.get("scope", "")
        query = f"{keywords} in:{scope}"

        # Getting API key
        token = self._get_api_key()
        auth = Auth.Token(token)
        github_connection = Github(auth=auth)

        # Searching on Github
        repositories = github_connection.search_repositories(query=query)

        # Saving search results
        for repo in repositories:

            record = record_transformer.repo_to_record(repo=repo)
            github_feed.add_update_record(record)

        github_feed.save()
        github_connection.close()

    def search(self, rerun: bool = False) -> None:
        """Run a search on GitHub"""

        if self.search_source.search_type == SearchType.API:
            github_feed = self.search_source.get_api_feed(
                review_manager=self.review_manager,
                source_identifier=self.source_identifier,
                update_only=(not rerun),
            )
            self._run_api_search(github_feed)

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""
        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
                unique_id_field="url",
            )

            return records
        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for GitHub"""
        return record

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Prepare the record based on GitHub"""

        if Fields.URL not in record.data:
            return record

        pattern = r"https?://github\.com/([A-Za-z0-9_.-]+)/([A-Za-z0-9_.-]+)$"
        match = re.match(pattern, record.data[Fields.URL].strip('"'))
        if match:  # Check whether record contains GitHub url

            # get API access
            token = self._get_api_key()
            auth = Auth.Token(token)
            github_connection = Github(auth=auth)

            repo = github_connection.get_repo(match.group(1) + "/" + match.group(2))
            new_record = record_transformer.repo_to_record(repo=repo)
            record.data.update(new_record.data)

        return record



#! /usr/bin/env python
"""Convenience functions to write excel files"""
from __future__ import annotations

import pandas as pd

from colrev.constants import Fields

FIELDS = [
    Fields.ID,
    Fields.ENTRYTYPE,
    Fields.TITLE,
    Fields.AUTHOR,
    Fields.YEAR,
    Fields.JOURNAL,
    Fields.BOOKTITLE,
    Fields.VOLUME,
    Fields.NUMBER,
    Fields.PAGES,
    Fields.DOI,
    Fields.URL,
    Fields.FILE,
]


def to_dataframe(*, records_dict: dict) -> pd.DataFrame:
    """Convert a records dict to a pandas DataFrame"""
    data = []
    for record_id in sorted(records_dict.keys()):
        record_dict = records_dict[record_id]
        row = {}
        for field in FIELDS:
            if field in record_dict:
                row[field] = record_dict[field]
            else:
                row[field] = ""
        data.append(row)
    return pd.DataFrame(data)


def write_file(*, records_dict: dict, filename: str) -> None:
    """Write an excel file from a records dict"""
    data_frame = to_dataframe(records_dict=records_dict)
    # pylint: disable=abstract-class-instantiated
    writer = pd.ExcelWriter(filename, engine="xlsxwriter")
    data_frame.to_excel(writer, index=False)

    worksheet = writer.sheets["Sheet1"]
    for i, column in enumerate(data_frame.columns):
        column_width = max(data_frame[column].astype(str).map(len).max(), len(column))
        column_width = max(min(column_width, 130), 10)
        worksheet.set_column(i, i, column_width)

    writer.close()



#! /usr/bin/env python
"""Exclude records with non-latin alphabets as a prep operation"""
from __future__ import annotations

import zope.interface
from alphabet_detector import AlphabetDetector
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields

# pylint: disable=duplicate-code


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class ExcludeNonLatinAlphabetsPrep:
    """Prepares records by excluding ones that have a non-latin alphabet
    (in the title, author, journal, or booktitle field)"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=True)

    source_correction_hint = "check with the developer"
    always_apply_changes = True
    alphabet_detector = AlphabetDetector()

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.prep_operation = prep_operation

    def _mostly_latin_alphabet(self, str_to_check: str) -> bool:
        assert len(str_to_check) != 0
        nr_latin = 0
        for character in str_to_check:
            if self.alphabet_detector.only_alphabet_chars(character, "LATIN"):
                nr_latin += 1
        return nr_latin / len(str_to_check) > 0.75

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
    ) -> colrev.record.record.Record:
        """Prepare the records by excluding records whose metadata is not in Latin alphabet"""

        if self.prep_operation.polish:
            return record

        # TB:D join or check independently?
        str_to_check = " ".join(
            [
                record.data.get(Fields.TITLE, ""),
            ]
        )

        if str_to_check and not self._mostly_latin_alphabet(str_to_check):
            record.prescreen_exclude(reason="non_latin_alphabet")

        return record



#! /usr/bin/env python
"""Template for a custom PDFPrep PackageEndpoint"""
from __future__ import annotations

import random

import zope.interface

import colrev.package_manager.interfaces
import colrev.package_manager.package_settings
import colrev.process.operation
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import RecordState


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PDFPrepInterface)
class CustomPDFPrep:
    """Class for custom pdf-prep scripts"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    def __init__(
        self,
        *,
        pdf_prep_operation: colrev.ops.pdf_prep.PDFPrep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)

    def prep_pdf(
        self,
        pdf_prep_operation: colrev.ops.pdf_prep.PDFPrep,  # pylint: disable=unused-argument
        record: colrev.record.record.Record,
        pad: int,  # pylint: disable=unused-argument
    ) -> colrev.record.record.Record:
        """Prepare the PDF"""

        if random.random() < 0.8:  # nosec
            record.add_field_provenance_note(
                key=Fields.FILE, note="custom_issue_detected"
            )
            record.set_status(RecordState.pdf_needs_manual_preparation)

        return record



#! /usr/bin/env python
"""Utility to transform crossref/doi.org items to records"""
from __future__ import annotations

import html
import re
from copy import deepcopy

import colrev.exceptions as colrev_exceptions
import colrev.record.record
import colrev.record.record_prep
from colrev.constants import Fields
from colrev.constants import FieldValues

# pylint: disable=duplicate-code

TAG_RE = re.compile(r"<[a-z/][^<>]{0,12}>")


def _get_year(*, item: dict) -> str:
    year = "-1"
    try:
        if "published-print" in item:
            date_parts = item["published-print"]["date-parts"]
        elif "published" in item:
            date_parts = item["published"]["date-parts"]
        elif "published-online" in item:
            date_parts = item["published-online"]["date-parts"]
        else:
            return ""
        year = str(date_parts[0][0])
    except KeyError:
        pass
    return year


# pylint: disable=colrev-missed-constant-usage
def _get_authors(*, item: dict) -> str:
    authors_strings = []
    for author in item.get("author", "NA"):
        a_string = ""
        if "family" in author:
            a_string += author["family"]
            if "given" in author:
                a_string += f", {author['given']}"
            authors_strings.append(a_string)
    return " and ".join(authors_strings).replace(",,", ",")


def _get_number(*, item: dict) -> str:
    if "journal-issue" in item:
        if "issue" in item["journal-issue"]:
            item["issue"] = item["journal-issue"]["issue"]
    return str(item.get("issue", ""))


def _get_fulltext(*, item: dict) -> str:
    fulltext_link = ""
    # Note : better use the doi-link resolution
    fulltext_link_l = [
        u["URL"] for u in item.get("link", []) if "pdf" in u["content-type"]
    ]
    if len(fulltext_link_l) == 1:
        fulltext_link = fulltext_link_l.pop()
    return fulltext_link


# pylint: disable=colrev-missed-constant-usage
def _item_to_record(*, item: dict) -> dict:
    # Note: the format differst between crossref and doi.org

    if isinstance(item["title"], list):
        item[Fields.TITLE] = str(item["title"][0])
    assert isinstance(item[Fields.TITLE], str)

    if isinstance(item.get("container-title", ""), list):
        if len(item["container-title"]) > 0:
            item["container-title"] = item["container-title"][0]
        else:
            item["container-title"] = ""
    assert isinstance(item.get("container-title", ""), str)

    item[Fields.ENTRYTYPE] = "misc"
    if item.get("type", "NA") == "journal-article":
        item[Fields.ENTRYTYPE] = "article"
        item[Fields.JOURNAL] = item.get("container-title", "")
    elif item.get("type", "NA") == "proceedings-article":
        item[Fields.ENTRYTYPE] = "inproceedings"
        item[Fields.BOOKTITLE] = item.get("container-title", "")
    elif item.get("type", "NA") == "book":
        item[Fields.BOOKTITLE] = "book"
        item[Fields.SERIES] = item.get("container-title", "")

    item[Fields.AUTHOR] = _get_authors(item=item)
    item[Fields.YEAR] = _get_year(item=item)
    item[Fields.VOLUME] = str(item.get(Fields.VOLUME, ""))
    item[Fields.NUMBER] = _get_number(item=item)
    item[Fields.PAGES] = item.get("page", "").replace("-", "--")
    item[Fields.CITED_BY] = item.get("is-referenced-by-count", "")
    item[Fields.DOI] = item.get("DOI", "").upper()
    item[Fields.FULLTEXT] = _get_fulltext(item=item)

    return item


def _flag_retracts(*, record_dict: dict) -> dict:
    if "update-to" in record_dict:
        for update_item in record_dict["update-to"]:
            if update_item["type"] == "retraction":
                record_dict[Fields.RETRACTED] = FieldValues.RETRACTED
    if "(retracted)" in record_dict.get(Fields.TITLE, "").lower():
        record_dict[Fields.RETRACTED] = FieldValues.RETRACTED
    return record_dict


def _format_fields(*, record_dict: dict) -> dict:
    for key, value in record_dict.items():
        record_dict[key] = str(value).replace("{", "").replace("}", "")
        # Note : some dois (and their provenance) contain html entities
        if key not in [
            Fields.AUTHOR,
            Fields.TITLE,
            Fields.JOURNAL,
            Fields.BOOKTITLE,
            Fields.ABSTRACT,
        ]:
            continue
        if not isinstance(value, str):
            continue
        value = value.replace("<scp>", "{")
        value = value.replace("</scp>", "}")
        value = html.unescape(value)
        value = re.sub(TAG_RE, " ", value)
        value = value.replace("\n", " ")
        value = re.sub(r"\s+", " ", value).rstrip().lstrip("▪ ")
        if key == Fields.ABSTRACT:
            if value.startswith("Abstract "):
                value = value[8:]
        record_dict[key] = value.lstrip().rstrip()

    return record_dict


def _set_forthcoming(*, record_dict: dict) -> dict:
    if not any(x in record_dict for x in ["published-print", "published"]) or not any(
        x in record_dict for x in [Fields.VOLUME, Fields.NUMBER]
    ):
        record_dict.update(year="forthcoming")
        if Fields.YEAR in record_dict:
            record_dict.update(published_online=record_dict[Fields.YEAR])
    return record_dict


def _remove_fields(*, record_dict: dict) -> dict:
    # Drop empty and non-supported fields
    supported_fields = [
        Fields.ENTRYTYPE,
        Fields.ID,
        Fields.TITLE,
        Fields.AUTHOR,
        Fields.YEAR,
        Fields.JOURNAL,
        Fields.BOOKTITLE,
        Fields.VOLUME,
        Fields.NUMBER,
        Fields.PAGES,
        Fields.DOI,
        Fields.FULLTEXT,
        Fields.ABSTRACT,
        "warning",
        Fields.LANGUAGE,
    ]
    record_dict = {
        k: v for k, v in record_dict.items() if k in supported_fields and v != ""
    }

    if (
        record_dict.get(Fields.ABSTRACT, "")
        == "No abstract is available for this article."
    ):
        del record_dict[Fields.ABSTRACT]

    return record_dict


def json_to_record(*, item: dict) -> colrev.record.record_prep.PrepRecord:
    """Convert a crossref item to a record dict"""

    try:
        record_dict = _item_to_record(item=deepcopy(item))
        record_dict = _set_forthcoming(record_dict=record_dict)
        record_dict = _flag_retracts(record_dict=record_dict)
        record_dict = _format_fields(record_dict=record_dict)
        record_dict = _remove_fields(record_dict=record_dict)
    except (IndexError, KeyError) as exc:
        raise colrev_exceptions.RecordNotParsableException(
            f"RecordNotParsableException: {exc}"
        ) from exc

    return colrev.record.record_prep.PrepRecord(record_dict)



#! /usr/bin/env python
"""Checker for inconsistent-content."""
from __future__ import annotations

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class InconsistentContentChecker:
    """The InconsistentContentChecker"""

    msg = DefectCodes.INCONSISTENT_CONTENT

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the inconsistent-content checks"""

        for key in [Fields.JOURNAL, Fields.BOOKTITLE, Fields.AUTHOR]:
            if key not in record.data or record.ignored_defect(
                key=key, defect=self.msg
            ):
                continue

            if record.masterdata_is_curated():
                continue

            if self._inconsistent_content(record=record, key=key):
                record.add_field_provenance_note(key=key, note=self.msg)
            else:
                record.remove_field_provenance_note(key=key, note=self.msg)

    def _inconsistent_content(
        self, *, record: colrev.record.record.Record, key: str
    ) -> bool:
        if key == Fields.JOURNAL:
            if Fields.JOURNAL in record.data and any(
                x in record.data[Fields.JOURNAL].lower()
                for x in ["conference", "workshop"]
            ):
                return True
        if key == Fields.BOOKTITLE:
            if Fields.BOOKTITLE in record.data and any(
                x in record.data[Fields.BOOKTITLE].lower() for x in [Fields.JOURNAL]
            ):
                return True

        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(InconsistentContentChecker(quality_model))



#! /usr/bin/env python
"""Utility to transform github repositories into records"""
from __future__ import annotations

import re

from github import Github
from github.GithubException import GithubException

import colrev.record.record
import colrev.record.record_prep
from colrev.constants import Fields


def _set_title(*, record_dict: dict, citation_data: str) -> None:
    """Set repository title"""
    title = re.search(r"^\s*title:\s*(.+)\s*$", citation_data, re.M)
    if title:
        record_dict[Fields.TITLE] = title.group(1).strip().replace('"', "")


def _set_authors(*, record_dict: dict, citation_data: str) -> None:
    """Set repository authors"""
    authors = re.findall(
        r'- family-names:\s*"(.+)"\s*\n\s*given-names:\s*"(.+)"',
        citation_data,
        re.M,
    )
    record_dict[Fields.AUTHOR] = (
        colrev.record.record_prep.PrepRecord.format_author_field(
            " and ".join([a[1].strip() + " " + a[0].strip() for a in authors])
        )
    )


def _set_url(*, record_dict: dict, citation_data: str) -> None:
    """Set repository URL"""
    url = re.search(r"^\s*url:\s*(.+)\s*$", citation_data, re.M)
    if url:
        record_dict[Fields.URL] = url.group(1).strip().replace('"', "")


def _set_year_and_release_date(*, record_dict: dict, citation_data: str) -> None:
    """Set release date"""
    release_date = re.search(r"^\s*date-released:\s*(.+)\s*$", citation_data, re.M)
    if release_date:
        record_dict[Fields.DATE] = release_date.group(1).strip()
        record_dict[Fields.YEAR] = record_dict[Fields.DATE][:4]


def _set_version(record_dict: dict, citation_data: str) -> None:
    """Set current software version"""
    version = re.search(r"^\s*version:\s*(.+)\s*$", citation_data, re.M)
    if version:
        record_dict[Fields.GITHUB_VERSION] = version.group(1).strip()


def _get_contributors(repo: Github.Repository.Repository) -> str:
    return " and ".join(
        [c.login for c in repo.get_contributors() if not c.login.endswith("[bot]")]
    )


def _set_license(record_dict: dict, repo: Github.Repository.Repository) -> None:
    try:
        record_dict[Fields.GITHUB_LICENSE] = repo.get_license().license.name
    except GithubException:
        pass


def _set_nr_contributors(record_dict: dict, repo: Github.Repository.Repository) -> None:
    record_dict[Fields.GITHUB_NR_CONTRIBUTORS] = 0  # empty repository
    try:
        record_dict[Fields.GITHUB_NR_CONTRIBUTORS] = repo.get_contributors().totalCount
    except GithubException:
        pass


def _set_nr_commits(record_dict: dict, repo: Github.Repository.Repository) -> None:
    record_dict[Fields.GITHUB_NR_COMMITS] = 0  # empty repository
    try:
        record_dict[Fields.GITHUB_NR_COMMITS] = repo.get_commits().totalCount
    except GithubException:
        pass


def _update_record_based_on_citation_cff(
    record_dict: dict, repo: Github.Repository.Repository
) -> dict:
    try:
        content = repo.get_contents("CITATION.cff")
        citation_data = content.decoded_content.decode("utf-8")

        _set_title(record_dict=record_dict, citation_data=citation_data)
        _set_authors(record_dict=record_dict, citation_data=citation_data)
        _set_year_and_release_date(record_dict=record_dict, citation_data=citation_data)
        _set_version(record_dict, citation_data)
        _set_url(record_dict=record_dict, citation_data=citation_data)

    except GithubException:
        record_dict = {}
    return record_dict


def repo_to_record(
    *, repo: Github.Repository.Repository
) -> colrev.record.record.Record:
    """Convert a GitHub repository to a record"""

    # Set default values
    record_dict = {
        Fields.ENTRYTYPE: "software",
        Fields.TITLE: repo.name,
        Fields.AUTHOR: _get_contributors(repo),
        Fields.YEAR: repo.updated_at.strftime("%Y"),
        Fields.DATE: repo.updated_at.strftime("%Y/%m/%d"),
        Fields.URL: repo.html_url,
        Fields.ABSTRACT: repo.description,
        Fields.GITHUB_LANGUAGE: repo.language,
    }

    _set_license(record_dict, repo)
    _set_nr_contributors(record_dict, repo)
    _set_nr_commits(record_dict, repo)

    # Use data from CITATION.cff file (if available)
    _update_record_based_on_citation_cff(record_dict, repo)

    record_dict = {k: v for k, v in record_dict.items() if v}

    return colrev.record.record.Record(data=record_dict)



#!/usr/bin/env python3
"""OSF API Query class for querying the OSF API."""
import json
import typing

import requests

from colrev.constants import Fields

# pylint: disable=colrev-missed-constant-usage


class OSFApiQuery:
    """Class for querying the OSF API."""

    base_url = "https://api.osf.io/v2/nodes/?"
    fields = ["id", "type", "title", "year", "description", "tags", "date_created"]

    def __init__(self, *, parameters: dict, api_key: str) -> None:
        self.api_key = api_key
        self.headers = {"Authorization": f"Bearer {self.api_key}"}
        self.params: typing.Dict[str, str] = {}
        self.next_url = None
        self.params = {
            key: value for key, value in parameters.items() if key in self.fields
        }

    def _create_record_dict(self, item: dict) -> dict:
        attributes = item["attributes"]
        year = attributes["date_created"]
        url = item["links"]
        relationships = item["relationships"]
        contributors = relationships["contributors"]
        links = contributors["links"]
        related = links["related"]
        record_dict = {
            Fields.ID: item["id"],
            Fields.ENTRYTYPE: "misc",
            Fields.AUTHOR: related["href"],
            Fields.TITLE: attributes["title"],
            Fields.ABSTRACT: attributes["description"],
            Fields.KEYWORDS: ", ".join(attributes["tags"]),
            Fields.YEAR: year[:4],
            Fields.URL: url["self"],
        }
        # Drop empty fields
        record_dict = {k: v for k, v in record_dict.items() if v}
        return record_dict

    def _query_api(self, url: str) -> str:
        """Creates the URL for the API call
        string url  Full URL to pass to API
        return string: Results from API"""

        response = requests.get(url, headers=self.headers, timeout=60)
        return response.text

    def retrieve_records(self) -> typing.List[dict]:
        """Call the API with the query parameters and return the results."""

        if self.next_url is None:
            url = self._build_query()
        else:
            url = self.next_url

        data = self._query_api(url)
        response = json.loads(data)

        # Set the next URL for the next query
        links = response["links"]
        self.next_url = links["next"]

        articles = response.get("data", [])
        return [self._create_record_dict(article) for article in articles]

    def overall(self) -> int:
        """Return the overall number of records."""

        url = self._build_query()
        data = self._query_api(url)
        response = json.loads(data)

        links = response["links"]
        return links["meta"]["total"]

    def _build_query(self) -> str:
        """Creates the URL for querying the API with support for nested filter parameters."""

        filters = []
        # Add in filters with the correct formatting
        for key, value in self.params.items():
            filters.append(f"filter[{key}]={value}")

        url = self.base_url + "&".join(filters)

        return url

    def pages_completed(self) -> bool:
        """Check if all pages have been completed"""
        return self.next_url is None



#!/usr/bin/env python3
"""Linter for CoLRev - direct status assignment"""
from __future__ import annotations

import typing

from astroid import nodes
from pylint import checkers
from pylint.checkers.utils import only_required_for_messages

from colrev.constants import Fields

if typing.TYPE_CHECKING:  # pragma: no cover
    from pylint.lint import PyLinter


class DirectStatusAssignmentChecker(checkers.BaseChecker):
    """DirectStatusAssignmentChecker"""

    name = "colrev-direct-status-assign"

    msgs = {
        "W0931": (
            "Direct status assignment",
            "colrev-direct-status-assign",
            "Emitted when a colrev-status is directly assigned "
            + "(instead of using record.set_stautus)",
        ),
    }

    @only_required_for_messages("colrev-direct-status-assign")
    def visit_assign(self, node: nodes.Assign) -> None:
        """
        Detect direct assignment of colrev_status.
        """

        if len(node.targets) != 1 or not hasattr(
            node.targets[0], "slice"
        ):  # pragma: no cover
            return

        if hasattr(node.targets[0].slice, "value"):
            if Fields.STATUS == node.targets[0].slice.value:
                self.add_message(self.name, node=node)
            return

    @only_required_for_messages("colrev-direct-status-assign")
    def visit_call(self, node: nodes.Assign) -> None:
        """
        Detect direct assignment of colrev_status.
        """
        if isinstance(node.func, nodes.Attribute) and node.func.attrname == "update":
            for keyword in node.keywords:
                if keyword.arg == "colrev_status":
                    self.add_message("colrev-direct-status-assign", node=node)


def register(linter: PyLinter) -> None:  # pragma: no cover
    """required method to auto register this checker"""

    linter.register_checker(DirectStatusAssignmentChecker(linter))



#! /usr/bin/env python
"""CoLRev debug prep operation."""
from __future__ import annotations

import logging
import time
import typing
from multiprocessing import Value

from requests.exceptions import ConnectionError as requests_ConnectionError

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.ops.prep
import colrev.process.operation
import colrev.record.record_prep
from colrev.constants import Colors
from colrev.constants import Fields

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.package_manager.interfaces
    import colrev.review_manager

# logging.getLogger("urllib3").setLevel(logging.ERROR)
logging.getLogger("requests_cache").setLevel(logging.ERROR)

PREP_COUNTER = Value("i", 0)


# pylint: disable=too-many-instance-attributes
class PrepDebug(colrev.ops.prep.Prep):
    """Debug prepare records (metadata)"""

    debug_ids: typing.List[str]
    commit_sha: str

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool,
        polish: bool,
    ) -> None:

        super().__init__(
            review_manager=review_manager,
            notify_state_transition_operation=notify_state_transition_operation,
            polish=polish,
            cpu=1,
        )

    # Overrides parent method
    def _print_diffs_for_debug(
        self,
        *,
        prior: colrev.record.record_prep.PrepRecord,
        preparation_record: colrev.record.record_prep.PrepRecord,
        prep_package_endpoint: colrev.package_manager.interfaces.PrepInterface,
    ) -> None:

        diffs = prior.get_diff(preparation_record)

        if diffs:
            change_report = (
                f"{prep_package_endpoint} changed:\n"
                f"{Colors.ORANGE}{self.review_manager.p_printer.pformat(diffs)}{Colors.END}\n"
            )

            self.review_manager.logger.info(change_report)
            self.review_manager.logger.info(
                "To correct errors in the endpoint,"
                " open an issue at "
                "https://github.com/CoLRev-Environment/colrev/issues"
            )
            self.review_manager.logger.info(
                "To correct potential errors at source,"
                f" {prep_package_endpoint.source_correction_hint}"
            )
        else:
            self.review_manager.logger.info(f"{prep_package_endpoint} changed: -")
            time.sleep(0.1)
        print("\n")

    def _retrieve_records_from_history(self, original_records: list[dict]) -> list:
        git_repo = self.review_manager.dataset.get_repo()

        if self.commit_sha == "":
            self.commit_sha = git_repo.git.rev_parse("HEAD")

        # get the commit_sha preceding commit_sha
        preceding_commit_sha = git_repo.git.rev_parse(self.commit_sha + "^")

        prior_records = next(
            self.review_manager.dataset.load_records_from_history(
                commit_sha=preceding_commit_sha
            )
        )

        original_records_ids = [r["ID"] for r in original_records]
        return [r for r in prior_records.values() if r["ID"] in original_records_ids]

    def _load_temp_prep_to_resume(self, prepare_data: dict) -> None:
        pass

    # overrides _load_prep_data
    def _load_prep_data(self) -> dict:

        records = []
        original_records = list(
            self.review_manager.dataset.read_next_record(
                conditions=[{Fields.ID: ID} for ID in self.debug_ids]
            )
        )
        records = self._retrieve_records_from_history(original_records)
        if len(records) == 0:
            prep_data = {"nr_tasks": 0, "PAD": 0, "items": []}
        else:
            prep_data = {
                "nr_tasks": len(self.debug_ids),
                "PAD": len(self.debug_ids),
                "items": records,
            }
        return prep_data

    @colrev.process.operation.Operation.decorate()
    def run_debug(self, *, debug_ids: str = "NA") -> None:
        """Preparation of records (main entrypoint)"""

        self.debug_ids = debug_ids.split(",")

        self.review_manager.logger.info("Start debug prep")
        self.review_manager.logger.info(
            "The debugger will replay the preparation procedures"
            " step-by-step, allow you to identify potential errors, trace them to "
            "their colrev_origin and correct them."
        )

        self.polish = input("Polish mode (y/n)?") == "y"
        self.commit_sha = input(
            "Commit in which the error occurred (press Enter to select the most recent commit):"
        )

        try:
            for i, prep_round in enumerate(
                self.review_manager.settings.prep.prep_rounds
            ):
                self._setup_prep_round(i=i, prep_round=prep_round)
                preparation_data = self._get_prep_data_tasks(prep_round)

                if len(preparation_data) == 0 and not self.temp_records.is_file():
                    self.review_manager.logger.info("No records to prepare.")
                    print()
                    return
                self.review_manager.logger.info("Loaded data. Start preparation.")
                for item in preparation_data:
                    record = self.prepare(item)
                    self.review_manager.logger.info(
                        f"Result:\n" f"{self.review_manager.p_printer.pformat(record)}"
                    )
        except requests_ConnectionError as exc:
            if "OSError(24, 'Too many open files" in str(exc):
                raise colrev_exceptions.ServiceNotAvailableException(
                    "Too many files opened (OSError, Errno24). "
                    "To use a smaller number of parallel processes, run colrev prep --cpu 1"
                ) from exc
            raise exc

        except OSError as exc:
            if 24 == exc.errno:
                raise colrev_exceptions.ServiceNotAvailableException(
                    "Too many files opened (OSError, Errno24). "
                    "To use a smaller number of parallel processes, run colrev prep --cpu 1"
                ) from exc
            raise exc



#! /usr/bin/env python
"""LocalIndex: sqlite."""
from __future__ import annotations

import sqlite3
import typing

import pandas as pd

import colrev.exceptions as colrev_exceptions
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import Filepaths
from colrev.constants import LocalIndexFields

# Note : records are indexed by id = hash(colrev_id)
# to ensure that the indexing-ids do not exceed limits
# such as the opensearch limit of 512 bytes.
# This enables efficient retrieval based on id=hash(colrev_id)
# but also search-based retrieval using only colrev_ids

# AUTHOR_INDEX = "author_index"
# AUTHOR_RECORD_INDEX = "author_record_index"
# CITATIONS_INDEX = "citations_index"

# # import binascii
# def _increment_hash(self, *, paper_hash: str) -> str:
#     plaintext = binascii.unhexlify(paper_hash)
#     # also, we'll want to know our length later on
#     plaintext_length = len(plaintext)
#     plaintext_number = int.from_bytes(plaintext, "big")

#     # recommendation: do not increment by 1
#     plaintext_number += 10
#     max_len_sha256 = 2**256
#     plaintext_number = plaintext_number % max_len_sha256

#     new_plaintext = plaintext_number.to_bytes(plaintext_length, "big")
#     new_hex = binascii.hexlify(new_plaintext)
#     # print(new_hex.decode("utf-8"))

#     return new_hex.decode("utf-8")


# pylint: disable=too-few-public-methods
class SQLiteIndex:
    """The SQLiteIndex class implements indexing and retrieval of records locally"""

    connection: sqlite3.Connection
    CREATE_TABLE_QUERY: str

    def __init__(
        self, *, index_name: str, index_keys: list, reinitialize: bool
    ) -> None:
        self.index_name = index_name
        self.index_keys = index_keys
        self.connection = sqlite3.connect(
            str(Filepaths.LOCAL_INDEX_SQLITE_FILE), timeout=90
        )
        self.connection.row_factory = self._dict_factory
        if reinitialize:
            self._reinitialize_db()

    def _dict_factory(self, cursor: sqlite3.Cursor, row: dict) -> dict:
        ret_dict = {}
        for idx, col in enumerate(cursor.description):
            ret_dict[col[0]] = row[idx]
        return ret_dict

    def _get_cursor(self) -> sqlite3.Cursor:
        return self.connection.cursor()

    def commit(self) -> None:
        """Commit changes to the SQLITE database"""
        if self.connection:
            self.connection.commit()

    def _reinitialize_db(self) -> None:
        """Reinitialize the SQLITE database"""

        cur = self._get_cursor()
        cur.execute(f"drop table if exists {self.index_name}")
        cur.execute(self.CREATE_TABLE_QUERY)
        if self.connection:
            self.connection.commit()

    def _get_record_from_row(self, row: dict) -> dict:

        records_dict = colrev.loader.load_utils.loads(
            load_string=row[LocalIndexFields.BIBTEX],
            implementation="bib",
            unique_id_field="ID",
        )

        retrieved_record = list(records_dict.values())[0]
        return retrieved_record


class SQLiteIndexRecord(SQLiteIndex):
    """The SQLiteIndexRecord class implements indexing and retrieval of records locally"""

    INDEX_NAME = "record_index"
    KEYS = [
        LocalIndexFields.ID,
        Fields.COLREV_ID,
        LocalIndexFields.CITATION_KEY,
        Fields.TITLE,
        Fields.ABSTRACT,
        Fields.FILE,
        LocalIndexFields.TEI,
        Fields.FULLTEXT,
        Fields.URL,
        Fields.DOI,
        LocalIndexFields.DBLP_KEY,  # Note : no dots in key names
        Fields.PDF_ID,
        LocalIndexFields.BIBTEX,
    ]

    GLOBAL_KEYS = [
        Fields.DOI,
        Fields.DBLP_KEY,
        Fields.PDF_ID,
        Fields.URL,
        Fields.COLREV_ID,
    ]

    CREATE_TABLE_QUERY = (
        f"CREATE TABLE {INDEX_NAME} (id TEXT PRIMARY KEY," + ",".join(KEYS[1:]) + ")"
    )

    SELECT_ALL_QUERY = f"SELECT * FROM {INDEX_NAME} WHERE"

    SELECT_KEY_QUERIES = {
        LocalIndexFields.ID: f"SELECT * FROM {INDEX_NAME} WHERE {LocalIndexFields.ID}=?",
        Fields.COLREV_ID: f"SELECT * FROM {INDEX_NAME} WHERE {Fields.COLREV_ID}=?",
        Fields.DOI: f"SELECT * FROM {INDEX_NAME} where {Fields.DOI}=?",
        Fields.DBLP_KEY: f"SELECT * FROM {INDEX_NAME} WHERE {Fields.DBLP_KEY}=?",
        Fields.PDF_ID: f"SELECT * FROM {INDEX_NAME} WHERE {Fields.PDF_ID}=?",
        Fields.URL: f"SELECT * FROM {INDEX_NAME} WHERE {Fields.URL}=?",
    }

    INSERT_QUERY = f"INSERT INTO {INDEX_NAME} VALUES(:{', :'.join(KEYS)})"

    UPDATE_RECORD_QUERY = f"""
            UPDATE {INDEX_NAME} SET
            {LocalIndexFields.BIBTEX}=?
            WHERE {LocalIndexFields.ID}=?"""

    def __init__(self, *, reinitialize: bool = False) -> None:
        super().__init__(
            index_name=self.INDEX_NAME,
            index_keys=self.KEYS,
            reinitialize=reinitialize,
        )

    def exists(
        self,
        *,
        local_index_id: str,
    ) -> bool:
        """Check if a record exists in the index"""
        cur = self._get_cursor()
        cur.execute(
            self.SELECT_KEY_QUERIES[LocalIndexFields.ID],
            (local_index_id,),
        )
        selected_row = cur.fetchone()
        if not selected_row:
            return False
        return True

    def insert(self, item: dict) -> None:
        """Insert a record into the index"""
        # May raise sqlite3.IntegrityError
        cur = self._get_cursor()
        cur.execute(self.INSERT_QUERY, item)
        self.commit()

    def get(
        self,
        *,
        key: str,
        value: str,
    ) -> dict:
        """Get a record from the index"""
        try:
            cur = self._get_cursor()
            cur.execute(self.SELECT_KEY_QUERIES[key], (value,))
            selected_row = cur.fetchone()
            if not selected_row:
                raise colrev_exceptions.RecordNotInIndexException(key)

            retrieved_record = {}
            retrieved_record = self._get_record_from_row(selected_row)
            if key != Fields.COLREV_ID and (
                key not in retrieved_record or value != retrieved_record[key]
            ):
                # Note: colrev-id collisions are handled separately
                raise colrev_exceptions.RecordNotInIndexException(key)

        except sqlite3.OperationalError as exc:  # pragma: no cover
            raise colrev_exceptions.RecordNotInIndexException(key) from exc

        # Handling collisions in colrev-ids
        if key == Fields.COLREV_ID:
            stored_colrev_id = colrev.record.record.Record(
                retrieved_record
            ).get_colrev_id()
            if value != stored_colrev_id:  # pragma: no cover
                print("Collisions (TODO):")
                print(stored_colrev_id)
                print(value)

                # print(
                #     [
                #         {k: v for k, v in x.items() if k != LocalIndexFields.BIBTEX}
                #         for x in stored_record
                #     ]
                # )
                # print(item)
                # to handle the collision:
                # print(f"Collision: {paper_hash}")
                # print(cid_to_index)
                # print(saved_record_cid)
                # print(saved_record)
                # paper_hash = self._increment_hash(paper_hash=paper_hash)
                # item[LocalIndexFields.ID] = paper_hash
                # continue in while-loop/try to insert...
                # pylint: disable=raise-missing-from
                raise NotImplementedError
        return retrieved_record

    def update(self, local_index_id: str, bibtex: str) -> None:
        """Update a record in the index"""
        cur = self._get_cursor()
        cur.execute(self.UPDATE_RECORD_QUERY, (bibtex, local_index_id))

    def search(self, query: str) -> list:
        """Search for records in the index"""
        cur = self._get_cursor()
        records_to_return = []

        selected_row = None
        cur.execute(f"{self.SELECT_ALL_QUERY} {query}")
        for row in cur.fetchall():
            selected_row = row
            retrieved_record_dict = self._get_record_from_row(selected_row)
            records_to_return.append(retrieved_record_dict)
        return records_to_return


class SQLiteIndexRankings(SQLiteIndex):
    """The SQLiteIndexRankings class implements indexing and retrieval
    of journal rankings locally"""

    INDEX_NAME = "rankings"
    KEYS: typing.List[str] = []

    CREATE_TABLE_QUERY = f"CREATE TABLE {INDEX_NAME} (id TEXT PRIMARY KEY)"
    SELECT_QUERY = f"SELECT * FROM {INDEX_NAME} WHERE journal_name = ?"

    def __init__(self, *, reinitialize: bool = False) -> None:
        super().__init__(
            index_name=self.INDEX_NAME,
            index_keys=self.KEYS,
            reinitialize=reinitialize,
        )

    def insert_df(self, data_frame: pd.DataFrame) -> None:
        """Insert a dataframe of journal rankings into the index"""
        conn = sqlite3.connect(str(Filepaths.LOCAL_INDEX_SQLITE_FILE))
        data_frame.to_sql(self.INDEX_NAME, conn, if_exists="replace", index=False)
        conn.commit()
        conn.close()

    def select(self, journal: str) -> list:
        """Select journal rankings from the index"""

        cur = self._get_cursor()
        cur.execute(self.SELECT_QUERY, (journal,))
        rankings = cur.fetchall()
        return rankings


class SQLiteIndexTOC(SQLiteIndex):
    """The SQLiteIndexTOC class implements indexing and retrieval of TOC items locally"""

    INDEX_NAME = "toc_index"
    KEYS: typing.List[str] = []

    CREATE_TABLE_QUERY = (
        f"CREATE TABLE {INDEX_NAME} (toc_key TEXT PRIMARY KEY, colrev_ids)"
    )

    SELECT_ALL_QUERY = f"SELECT * FROM {INDEX_NAME} WHERE "

    SELECT_KEY_QUERY = {
        LocalIndexFields.TOC_KEY: f"SELECT * FROM {INDEX_NAME} WHERE {LocalIndexFields.TOC_KEY}=?",
    }

    INSERT_MANY_QUERY = f"INSERT INTO {INDEX_NAME} VALUES(?, ?)"

    def __init__(self, *, reinitialize: bool = False) -> None:
        super().__init__(
            index_name=self.INDEX_NAME,
            index_keys=self.KEYS,
            reinitialize=reinitialize,
        )

    def exists(self, toc_item: str) -> bool:
        """Check if TOC item exists in the index"""
        cur = self._get_cursor()
        cur.execute(
            self.SELECT_KEY_QUERY[LocalIndexFields.TOC_KEY],
            (toc_item,),
        )
        selected_row = cur.fetchone()
        if not selected_row:
            return False
        return True

    def get_toc_items(self, toc_key: str = "", partial_toc_key: str = "") -> list:
        """Get TOC items from the index"""
        if partial_toc_key != "":
            query = f"{self.SELECT_ALL_QUERY} {LocalIndexFields.TOC_KEY} LIKE ?"
            argument = partial_toc_key + "%"
        elif toc_key != "":
            query = (
                f"SELECT * FROM {self.INDEX_NAME} WHERE {LocalIndexFields.TOC_KEY}=?"
            )
            argument = toc_key
        else:
            raise NotImplementedError

        try:
            cur = self._get_cursor()
            cur.execute(query, (argument,))
            results = cur.fetchall()

        except sqlite3.OperationalError as exc:  # pragma: no cover
            raise colrev_exceptions.RecordNotInIndexException(toc_key) from exc
        except AttributeError as exc:
            raise colrev_exceptions.RecordNotInIndexException(toc_key) from exc

        # toc_items = results.get("colrev_ids", "").split(";")  # type: ignore

        toc_items = [x["colrev_ids"].split(";") for x in results]
        toc_items = [item for sublist in toc_items for item in sublist]

        return toc_items

    def add(self, toc_to_index: dict) -> None:
        """Add TOC items to the index"""
        list_to_add = list((k, v) for k, v in toc_to_index.items() if v != "DROPPED")
        cur = self._get_cursor()
        try:
            cur.executemany(self.INSERT_MANY_QUERY, list_to_add)
        except sqlite3.IntegrityError as exc:  # pragma: no cover
            print(exc)
        finally:
            self.commit()



#! /usr/bin/env python
"""Convenience functions to write ris files"""
from __future__ import annotations

from colrev.constants import ENTRYTYPES
from colrev.constants import Fields

RECORDS_FIELD_ORDER = [
    "TY",  # Must be the first field
    "TI",
    "AU",
    "PY",
    "SP",
    "EP",
    "JO",
    "VL",
    "ER",  # Must be the last field
]

ENTRYTYPE_MAP = {
    ENTRYTYPES.ARTICLE: "JOUR",
    ENTRYTYPES.BOOK: "BOOK",
    ENTRYTYPES.INPROCEEDINGS: "CONF",
    ENTRYTYPES.MISC: "GEN",
    ENTRYTYPES.PHDTHESIS: "THES",
    ENTRYTYPES.TECHREPORT: "RPRT",
    ENTRYTYPES.UNPUBLISHED: "UNPB",
}


RECORD_FIELD_MAP = {
    "TI": Fields.TITLE,
    "PY": Fields.YEAR,
    "JO": Fields.JOURNAL,
    "VL": Fields.VOLUME,
}


def to_string(*, records_dict: dict) -> str:
    """Convert a records dict to a RIS string"""
    ris_str = ""
    for record_id in sorted(records_dict.keys()):
        record_dict = records_dict[record_id]
        for field in RECORDS_FIELD_ORDER:
            if field == "TY":
                entrytype = ENTRYTYPE_MAP[record_dict[Fields.ENTRYTYPE]]
                ris_str += f"TY  - {entrytype}\n"

            elif field in RECORD_FIELD_MAP:
                ris_str += f"{field}  - {record_dict[RECORD_FIELD_MAP[field]]}\n"
            elif field == Fields.PAGES:
                if Fields.PAGES in record_dict:
                    pages = record_dict[Fields.PAGES]
                    if "-" in pages:
                        start, end = pages.split("-")
                        ris_str += f"SP  - {start}\n"
                        ris_str += f"EP  - {end}\n"
                    else:
                        ris_str += f"SP  - {pages}\n"

            elif field == "AU":
                for author in record_dict[Fields.AUTHOR].split(" and "):
                    ris_str += f"AU  - {author}\n"

            elif field == "ER":
                ris_str += "ER  -\n\n"
                break  # must be last field

            else:
                pass  # field not (yet) supported

    return ris_str


def write_file(*, records_dict: dict, filename: str) -> None:
    """Write a RIS file from a records dict"""
    ris_str = to_string(records_dict=records_dict)
    with open(filename, "w", encoding="utf-8") as file:
        file.write(ris_str)



#! /usr/bin/env python
"""CoLRev package."""
from __future__ import annotations

import importlib.util
from importlib.metadata import distribution
from importlib.metadata import distributions
from importlib.metadata import PackageNotFoundError
from pathlib import Path
from typing import Any

import zope.interface.exceptions
from zope.interface.verify import verifyClass

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.process.operation
import colrev.record.record
import colrev.settings
from colrev.constants import EndpointType

# Inspiration for package descriptions:
# https://github.com/rstudio/reticulate/blob/
# 9ebca7ecc028549dadb3d51d2184f9850f6f9f9d/DESCRIPTION


ENDPOINT_OVERVIEW = colrev.package_manager.interfaces.ENDPOINT_OVERVIEW


class Package:
    """A Python package for CoLRev"""

    def __init__(self, package_identifier: str) -> None:
        try:
            self.package = distribution(package_identifier)
        except PackageNotFoundError as exc:
            # Note: The distribution(package_identifier)
            # does not seem to work reliably (across matrix tests on GitHub Actions)
            for dist in distributions():
                if dist.metadata["Name"] == package_identifier:
                    self.package = dist
                    break
            else:
                raise colrev_exceptions.MissingDependencyError(
                    f"Package {package_identifier} not found"
                ) from exc

        if not self.package.files:
            raise colrev_exceptions.MissingDependencyError(
                f"Package {package_identifier} not a CoLRev package " "(no files found)"
            )

        package_path = self.package.files[0].locate()
        self.package_dir = Path(package_path).parent

        # Note: The metadata() function does not work reliably
        # package_metadata = metadata(package_identifier)
        self.name = self.package.metadata["Name"]
        self.version = self.package.metadata["Version"]

    def has_endpoint(self, endpoint_type: EndpointType) -> bool:
        """Check if the package has a specific endpoint type"""

        return endpoint_type.value in [e.name for e in self.package.entry_points]

    def get_endpoint(self, endpoint_type: EndpointType) -> str:
        """Get the endpoint for a package type"""

        if endpoint_type.value not in [e.name for e in self.package.entry_points]:
            raise colrev_exceptions.MissingDependencyError(
                f"Package {self.name} does not have a {endpoint_type} endpoint"
            )

        return [e for e in self.package.entry_points if e.name == endpoint_type.value][
            0
        ].value

    def _endpoint_verified(
        self, endpoint_class: Any, endpoint_type: EndpointType, identifier: str
    ) -> bool:
        interface_definition = ENDPOINT_OVERVIEW[endpoint_type]["import_name"]
        try:
            verifyClass(interface_definition, endpoint_class)  # type: ignore
            return True
        except zope.interface.exceptions.BrokenImplementation as exc:
            print(f"Error registering endpoint {identifier}: {exc}")
        return False

    def get_endpoint_class(self, package_type: EndpointType) -> Any:
        """Get the endpoint class for a package type"""
        if not self.has_endpoint(package_type):
            raise colrev_exceptions.MissingDependencyError(
                f"Package {self.name} does not have a {package_type} endpoint"
            )

        endpoint_path = self.get_endpoint(package_type)
        module_name, class_name = endpoint_path.split(":")
        module = importlib.import_module(module_name)
        cls = getattr(module, class_name)
        if not self._endpoint_verified(cls, package_type, self.name):
            raise colrev_exceptions.MissingDependencyError(
                f"Endpoint {class_name} in {module_name} "
                f"does not implement the {package_type} interface"
            )
        return cls

    def add_to_type_identifier_endpoint_dict(
        self, type_identifier_endpoint_dict: dict
    ) -> None:
        """Add the package to the type_identifier_endpoint_dict dict"""

        for endpoint_type in EndpointType:
            if self.has_endpoint(endpoint_type):
                type_identifier_endpoint_dict[endpoint_type][self.name] = (
                    self.get_endpoint(endpoint_type)
                )



#! /usr/bin/env python
"""Convenience functions to load files (BiBTeX, RIS, CSV, etc.)"""
import logging
import typing
from pathlib import Path

from colrev.constants import ENTRYTYPES
from colrev.constants import Fields


# pylint: disable=too-many-arguments


class Loader:
    """Loader class"""

    def __init__(
        self,
        *,
        filename: Path,
        entrytype_setter: typing.Callable,
        field_mapper: typing.Callable,
        id_labeler: typing.Callable,
        unique_id_field: str,
        logger: logging.Logger,
    ):
        self.filename = filename
        self.unique_id_field = unique_id_field
        assert id_labeler is not None or unique_id_field != ""
        self.id_labeler = id_labeler
        self.entrytype_setter = entrytype_setter
        self.field_mapper = field_mapper

        self.logger = logger

    def _set_ids(self, records_list: list) -> None:
        if self.unique_id_field == "INCREMENTAL":
            for next_id, record_dict in enumerate(records_list, 1):
                record_dict[Fields.ID] = str(next_id).zfill(6)

        elif self.unique_id_field != "":
            for record_dict in records_list:
                record_dict[Fields.ID] = record_dict[self.unique_id_field]
        else:
            self.id_labeler(records_list)  # type: ignore

        assert all(
            Fields.ID in record_dict for record_dict in records_list
        ), "ID not set in all records"
        unique_ids = {record_dict[Fields.ID] for record_dict in records_list}
        assert len(unique_ids) == len(records_list), "ID is not unique in records"

    def _set_entrytypes(self, records_dict: dict) -> None:
        for r_dict_val in records_dict.values():
            self.entrytype_setter(r_dict_val)
        assert all(
            Fields.ENTRYTYPE in r for r in records_dict.values()
        ), "ENTRYTYPE not set in all records"
        invalid_entrytypes = [
            r[Fields.ENTRYTYPE]
            for r in records_dict.values()
            if r[Fields.ENTRYTYPE] not in ENTRYTYPES.get_all()
        ]
        assert (
            len(invalid_entrytypes) == 0
        ), f"Invalid ENTRYTYPE in some records: {invalid_entrytypes}"

    def _set_fields(self, records_dict: dict) -> None:
        for record_dict in records_dict.values():
            self.field_mapper(record_dict)

        # assert all(
        #     mandatory_field in record_dict
        #     for record_dict in records_dict.values()
        #     for mandatory_field in [Fields.AUTHOR, Fields.TITLE, Fields.YEAR]
        # ), "Mandatory field not set in all records"

        error_field_list = [
            field if any(c in field for c in [" ", ";"]) else None
            for record_dict in records_dict.values()
            for field in record_dict.keys()
        ]

        error_fields = {field for field in error_field_list if field is not None}

        if any(error_fields):
            error_cases = [
                r
                for r in records_dict.values()
                if any(error_field in r for error_field in error_fields)
            ]
            self.logger.error(
                f"Record contains invalid keys: {error_fields},\n record: {error_cases}"
            )

    def load_records_list(self) -> list:
        """The load_records_list must be implemented by the inheriting class
        (e.g., for ris/bib/...)"""
        raise NotImplementedError  # pragma: no cover

    def load(self) -> dict:
        """Load table entries from the source"""

        records_list = self.load_records_list()
        self._set_ids(records_list)
        records_dict = {str(r[Fields.ID]): r for r in records_list}
        self._set_entrytypes(records_dict)
        self._set_fields(records_dict)

        return records_dict



#! /usr/bin/env python
"""Functionality for PDF handling."""
from __future__ import annotations

import logging
import os
import re
import tempfile
import typing
import unicodedata
from pathlib import Path

import imagehash
import pymupdf
from PIL import Image

import colrev.env.utils
import colrev.exceptions as colrev_exceptions
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import Fields


class PDFRecord(colrev.record.record.Record):
    """The PDFRecord class provides a range of convenience functions for PDF handling"""

    def __init__(self, data: dict, path: Path) -> None:
        self.data = data
        """Dictionary containing the record data"""

        self.path = path
        """Path to the repository (record.data[Fields.File] is relative to path)"""

        super().__init__(data=data)

    def _get_path(self) -> Path:
        if Fields.FILE not in self.data:
            raise colrev_exceptions.InvalidPDFException(path=self.data[Fields.ID])

        pdf_path = (self.path / Path(self.data[Fields.FILE])).absolute()

        if not pdf_path.is_file():
            raise colrev_exceptions.InvalidPDFException(path=pdf_path)

        if 0 == os.path.getsize(pdf_path):
            logging.error("%sPDF with size 0: %s %s", Colors.RED, pdf_path, Colors.END)
            raise colrev_exceptions.InvalidPDFException(path=pdf_path)
        return pdf_path

    def _fix_text_encoding_issues(self, text: str) -> str:
        # Correct common incorrect encoding issues
        # Handle cases where the diacritical mark is on a different line or misplaced
        text = re.sub(r"(´)\n([aeiou])", r"\1\2", text)
        text = re.sub(r"(ˆ)\n([aeiou])", r"\1\2", text)

        # Replace sequences of characters back with the correct accented characters
        text = (
            text.replace("´e", "é")
            .replace("´a", "á")
            .replace("´i", "í")
            .replace("´o", "ó")
            .replace("´u", "ú")
        )
        text = (
            text.replace("ˆo", "ô")
            .replace("ˆa", "â")
            .replace("ˆe", "ê")
            .replace("ˆi", "î")
            .replace("ˆu", "û")
        )

        # Normalize the text to combine characters and their diacritics properly
        text = unicodedata.normalize("NFKC", text)

        return text

    def extract_text_by_page(
        self,
        *,
        pages: typing.Optional[list] = None,
    ) -> str:
        """Extract the text from the PDF for a given number of pages"""
        pdf_path = self._get_path()
        text_list: list = []
        with pymupdf.open(pdf_path) as doc:
            for i, page in enumerate(doc):
                if pages is None or i in pages:
                    text = page.get_text()
                    text_list.append(text)

        text_all = "".join(text_list)
        return self._fix_text_encoding_issues(text_all)

    def set_nr_pages_in_pdf(self) -> None:
        """Set the pages_in_file field based on the PDF"""
        pdf_path = self._get_path()

        with pymupdf.open(pdf_path) as doc:
            pages_in_file = doc.page_count
        self.data[Fields.NR_PAGES_IN_FILE] = pages_in_file

    def set_text_from_pdf(self) -> None:
        """Set the text_from_pdf field based on the PDF"""
        self.data[Fields.TEXT_FROM_PDF] = ""
        self.set_nr_pages_in_pdf()
        text = self.extract_text_by_page(pages=[0, 1, 2])
        text_from_pdf = text.replace("\n", " ").replace("\x0c", "")
        self.data[Fields.TEXT_FROM_PDF] = text_from_pdf

    @classmethod
    def extract_pages_from_pdf(
        cls,
        *,
        pages: list,
        pdf_path: Path,
        save_to_path: typing.Optional[Path] = None,
    ) -> None:  # pragma: no cover
        """Extract pages from the PDF"""
        doc = pymupdf.Document(pdf_path)
        all_pages_list = list(range(doc.page_count))

        if save_to_path:
            save_doc = pymupdf.Document()
            for page in pages:
                save_doc.insert_pdf(doc, from_page=page, to_page=page)
            save_doc.save(save_to_path / Path(pdf_path).name)
            save_doc.close()

        saved_pdf_pages = []

        for page in pages:
            all_pages_list.remove(page)
            saved_pdf_pages.append(page)

        doc.select(all_pages_list)
        # pylint: disable=no-member
        doc.save(pdf_path, incremental=True, encryption=pymupdf.PDF_ENCRYPT_KEEP)
        doc.close()

    def extract_pages(
        self,
        *,
        pages: list,
        save_to_path: typing.Optional[Path] = None,
    ) -> None:  # pragma: no cover
        """Extract pages from the PDF"""
        pdf_path = self._get_path()
        self.extract_pages_from_pdf(
            pages=pages, pdf_path=pdf_path, save_to_path=save_to_path
        )

    def get_pdf_hash(self, *, page_nr: int, hash_size: int = 32) -> str:
        """Get the PDF image hash"""
        assert page_nr > 0
        assert hash_size in [16, 32]

        pdf_path = self._get_path()

        with tempfile.NamedTemporaryFile(suffix=".png") as temp_file:
            file_name = temp_file.name
            try:
                doc: pymupdf.Document = pymupdf.open(pdf_path)
                # Starting with page 1
                for page_no, page in enumerate(doc, 1):
                    if page_no == page_nr:
                        pix = page.get_pixmap(dpi=200)
                        pix.save(file_name)  # store image as a PNG
                        with Image.open(file_name) as img:
                            average_hash = imagehash.average_hash(
                                img, hash_size=hash_size
                            )
                            average_hash_str = str(average_hash).replace("\n", "")
                            if len(average_hash_str) * "0" == average_hash_str:
                                raise colrev_exceptions.PDFHashError(path=pdf_path)
                            return average_hash_str
                # Page not found
                raise colrev_exceptions.PDFHashError(path=pdf_path)  # pragma: no cover
            except pymupdf.FileDataError as exc:
                raise colrev_exceptions.InvalidPDFException(path=pdf_path) from exc
            except RuntimeError as exc:
                raise colrev_exceptions.PDFHashError(path=pdf_path) from exc



#!/usr/bin/env python3
"""Scripts printing information for the colrev show command"""
import platform
from pathlib import Path

import colrev.ops.check
import colrev.process.operation
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import RecordState


def print_sample(review_manager: colrev.review_manager.ReviewManager) -> None:
    """Print the sample on cli"""

    colrev.ops.check.CheckOperation(review_manager)
    records = review_manager.dataset.load_records_dict()
    sample = [
        r
        for r in records.values()
        if r[Fields.STATUS]
        in [
            RecordState.rev_synthesized,
            RecordState.rev_included,
        ]
    ]
    if 0 == len(sample):
        print("No records included in sample (yet)")

    for sample_r in sample:
        colrev.record.record.Record(sample_r).print_citation_format()


def print_venv_notes() -> None:
    """Print the virtual environment details on cli"""

    current_platform = platform.system()
    if current_platform == "Linux":
        print("Detected platform: Linux")
        if not Path("venv").is_dir():
            print("To create virtualenv, run")
            print(f"  {Colors.ORANGE}python3 -m venv venv{Colors.END}")
        print("To activate virtualenv, run")
        print(f"  {Colors.ORANGE}source venv/bin/activate{Colors.END}")
        print("To install colrev/colrev, run")
        print(f"  {Colors.ORANGE}python -m pip install colrev{Colors.END}")
        print("To deactivate virtualenv, run")
        print(f"  {Colors.ORANGE}deactivate{Colors.END}")
    elif current_platform == "Darwin":
        print("Detected platform: MacOS")
        if not Path("venv").is_dir():
            print("To create virtualenv, run")
            print(f"  {Colors.ORANGE}python3 -m venv venv{Colors.END}")
        print("To activate virtualenv, run")
        print(f"  {Colors.ORANGE}source venv/bin/activate{Colors.END}")
        print("To install colrev/colrev, run")
        print(f"  {Colors.ORANGE}python -m pip install colrev{Colors.END}")
        print("To deactivate virtualenv, run")
        print(f"  {Colors.ORANGE}deactivate{Colors.END}")
    elif current_platform == "Windows":
        print("Detected platform: Windows")
        if not Path("venv").is_dir():
            print("To create virtualenv, run")
            print(f"  {Colors.ORANGE}python -m venv venv{Colors.END}")
        print("To activate virtualenv, run")
        print(f"  {Colors.ORANGE}venv\\Scripts\\Activate.ps1{Colors.END}")
        print("To install colrev/colrev, run")
        print(f"  {Colors.ORANGE}python -m pip install colrev{Colors.END}")
        print("To deactivate virtualenv, run")
        print(f"  {Colors.ORANGE}deactivate{Colors.END}")
    else:
        print(
            "Platform not detected... "
            "cannot provide infos in how to activate virtualenv"
        )



#! /usr/bin/env python
"""Checker for inconsistent-with-doi-metadata."""
from __future__ import annotations

from rapidfuzz import fuzz

import colrev.exceptions as colrev_exceptions
import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields
from colrev.constants import FieldValues
from colrev.packages.crossref.src import crossref_api

# pylint: disable=too-few-public-methods


class InconsistentWithDOIMetadataChecker:
    """The InconsistentWithDOIMetadataChecker"""

    msg = DefectCodes.INCONSISTENT_WITH_DOI_METADATA
    _fields_to_check = [
        Fields.AUTHOR,
        Fields.TITLE,
        Fields.JOURNAL,
        Fields.YEAR,
        Fields.VOLUME,
        Fields.NUMBER,
    ]

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

        self.crossref_api = crossref_api.CrossrefAPI(params={})

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the inconsistent-with-doi-metadata checks"""

        if Fields.DOI not in record.data or record.ignored_defect(
            key=Fields.DOI, defect=self.msg
        ):
            return

        if "md_curated.bib" in record.get_field_provenance_source(Fields.DOI):
            return

        if self._doi_metadata_conflicts(record=record):
            record.add_field_provenance_note(key=Fields.DOI, note=self.msg)
        else:
            record.remove_field_provenance_note(key=Fields.DOI, note=self.msg)

    def _doi_metadata_conflicts(self, *, record: colrev.record.record.Record) -> bool:
        record_copy = record.copy_prep_rec()

        try:
            crossref_md = self.crossref_api.query_doi(doi=record_copy.data[Fields.DOI])

            for key in crossref_md.data.keys():
                if key not in self._fields_to_check:
                    continue
                if key not in record.data:
                    continue
                if record.data[key] == FieldValues.UNKNOWN:
                    continue
                if key not in [Fields.AUTHOR, Fields.TITLE, Fields.JOURNAL]:
                    continue
                if len(crossref_md.data[key]) < 5 or len(record.data[key]) < 5:
                    continue
                if (
                    fuzz.partial_ratio(
                        record.data[key].lower(), crossref_md.data[key].lower()
                    )
                    < 60
                ):
                    return True

        except (
            colrev_exceptions.RecordNotFoundInPrepSourceException,
            colrev_exceptions.RecordNotParsableException,
        ):
            return False
        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(InconsistentWithDOIMetadataChecker(quality_model))



#! /usr/bin/env python
"""Convenience functions to load nbib files"""
from __future__ import annotations

import logging
import re
import typing
from pathlib import Path

import colrev.loader.loader


# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes


class NextLine(Exception):
    """NextLineException"""


class ParseError(Exception):
    """Parsing error"""


class NBIBLoader(colrev.loader.loader.Loader):
    """Loads nbib files"""

    PATTERN = r"^[A-Z]{2,4}( ){1,2}- "

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        filename: Path,
        entrytype_setter: typing.Callable,
        field_mapper: typing.Callable,
        id_labeler: typing.Callable,
        unique_id_field: str = "",
        logger: logging.Logger = logging.getLogger(__name__),
    ):
        super().__init__(
            filename=filename,
            id_labeler=id_labeler,
            unique_id_field=unique_id_field,
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=logger,
        )

        self.current: dict = {}
        self.pattern = re.compile(self.PATTERN)

    @classmethod
    def get_nr_records(cls, filename: Path) -> int:
        """Get the number of records in the file"""
        count = 0
        with open(filename, encoding="utf-8") as file:
            for line in file:
                if line.startswith("TI "):
                    count += 1
        return count

    def _get_tag(self, line: str) -> str:
        """Get the tag from a line in the NBIB file."""
        return line[: line.find(" - ")].rstrip()

    def _get_content(self, line: str) -> str:
        """Get the content from a line"""
        return line[line.find(" - ") + 2 :].strip()

    def _add_tag(self, tag: str, line: str) -> None:
        new_value = self._get_content(line)

        if tag not in self.current:
            self.current[tag] = new_value
        elif isinstance(self.current[tag], str):
            self.current[tag] = [self.current[tag], new_value]
        elif isinstance(self.current[tag], list):
            self.current[tag].append(new_value)

    def _parse_tag(self, line: str) -> dict:
        tag = self._get_tag(line)

        if tag.strip() == "":
            return self.current

        self._add_tag(tag, line)
        raise NextLine

    def _parse_lines(self, lines: list) -> typing.Iterator[dict]:
        for line in lines:
            try:
                yield self._parse_tag(line)
                self.current = {}
            except NextLine:
                continue

    def load_records_list(self) -> list:
        """Loads nbib entries"""

        # based on
        # https://github.com/MrTango/rispy/blob/main/rispy/parser.py
        # Note: skip-tags and unknown-tags can be handled
        # between load_nbib_entries and convert_to_records.

        text = self.filename.read_text(encoding="utf-8")
        # clean_text?
        lines = text.split("\n")
        records_list = list(r for r in self._parse_lines(lines) if r)
        return records_list



#!/usr/bin/env python3
"""Linter for CoLRev - records variable naming convention"""
from __future__ import annotations

import typing

from astroid import nodes
from pylint import checkers
from pylint.checkers.utils import only_required_for_messages


if typing.TYPE_CHECKING:  # pragma: no cover
    from pylint.lint import PyLinter


class RecordsVariableNamingConventionChecker(checkers.BaseChecker):
    """RecordsVariableNamingConvention"""

    name = "colrev-records-variable-naming-convention"

    msgs = {
        "W0934": (
            "Records variable not named according to convention",
            "colrev-records-variable-naming-convention",
            "Emitted when the records variable is not named according to convention",
        ),
    }

    @only_required_for_messages("colrev-records-variable-naming-convention")
    def visit_assign(self, node: nodes.Assign) -> None:
        """
        Detect colrev-records-variable-naming-convention.
        """

        if len(node.targets) != 1:  # pragma: no cover
            return

        assigned = node.value

        if (
            not hasattr(assigned, "func")
            or not isinstance(assigned.func, nodes.Attribute)
            or assigned.func.attrname != "load_records_dict"
            or len(node.targets) != 1
            or not hasattr(node.targets[0], "name")
        ):
            return

        if hasattr(assigned, "keywords"):
            for keyword in assigned.keywords:
                if keyword.arg == "header_only":
                    if keyword.value.value is True:
                        if node.targets[0].name != "records_headers":
                            self.add_message(self.name, node=node)
                    else:
                        if node.targets[0].name != "records":
                            self.add_message(self.name, node=node)
                    return

        if node.targets[0].name != "records":
            self.add_message(self.name, node=node)


def register(linter: PyLinter) -> None:  # pragma: no cover
    """required method to auto register this checker"""

    linter.register_checker(RecordsVariableNamingConventionChecker(linter))



#! /usr/bin/env python
"""Prescreen based on specified scope"""
from __future__ import annotations

import typing

import zope.interface
from pydantic import BaseModel
from pydantic import Field

import colrev.env.language_service
import colrev.env.local_index
import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import RecordState


# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code

# to check: https://asistdl.onlinelibrary.wiley.com/doi/full/10.1002/asi.24816


class ScopePrescreenSettings(
    colrev.package_manager.package_settings.DefaultSettings, BaseModel
):
    """Settings for ScopePrescreen"""

    # pylint: disable=invalid-name
    # pylint: disable=too-many-instance-attributes

    endpoint: str
    ExcludePredatoryJournals: bool
    TimeScopeFrom: typing.Optional[int]
    TimeScopeTo: typing.Optional[int]
    LanguageScope: typing.Optional[list]
    ExcludeComplementaryMaterials: typing.Optional[bool]
    OutletInclusionScope: typing.Optional[dict]
    OutletExclusionScope: typing.Optional[dict]
    ENTRYTYPEScope: typing.Optional[list]
    RequireRankedJournals: typing.Optional[list]

    _details = {
        "TimeScopeFrom": {
            "tooltip": "Lower bound for the time scope",
            "min": 1900,
            "max": 2050,
        },
        "TimeScopeTo": {
            "tooltip": "Upper bound for the time scope",
            "min": 1900,
            "max": 2050,
        },
        "LanguageScope": {"tooltip": "Language scope"},
        "ExcludeComplementaryMaterials": {
            "tooltip": "Whether complementary materials (coverpages etc.) are excluded"
        },
        "OutletInclusionScope": {
            "tooltip": "Particular outlets that should be included (exclusively)"
        },
        "OutletExclusionScope": {
            "tooltip": "Particular outlets that should be excluded"
        },
        "ExcludePredatoryJournals": {"tooltip": "Exclude predatory journals"},
        "ENTRYTYPEScope": {
            "tooltip": "Particular ENTRYTYPEs that should be included (exclusively)"
        },
    }


@zope.interface.implementer(colrev.package_manager.interfaces.PrescreenInterface)
class ScopePrescreen:
    """Rule-based prescreen (scope)"""

    settings: ScopePrescreenSettings
    ci_supported: bool = Field(default=True)

    settings_class = ScopePrescreenSettings

    def __init__(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        settings: dict,
    ) -> None:
        if "TimeScopeFrom" in settings:
            settings["TimeScopeFrom"] = int(settings["TimeScopeFrom"])
            assert settings["TimeScopeFrom"] > 1900
            assert settings["TimeScopeFrom"] < 2100
        if "TimeScopeTo" in settings:
            settings["TimeScopeTo"] = int(settings["TimeScopeTo"])
            assert settings["TimeScopeTo"] > 1900
            assert settings["TimeScopeTo"] < 2100
        if "LanguageScope" in settings:
            self.language_service = colrev.env.language_service.LanguageService()
            try:
                self.language_service.validate_iso_639_3_language_codes(
                    lang_code_list=settings["LanguageScope"]
                )
            except colrev_exceptions.InvalidLanguageCodeException as exc:
                raise colrev_exceptions.InvalidSettingsError(
                    msg=f"Invalid LanguageScope in scope_prescreen: {settings['LanguageScope']}"
                    + " (should be iso_639_3 language code)",
                    fix_per_upgrade=False,
                ) from exc

        if "ExcludePredatoryJournals" not in settings:
            settings["ExcludePredatoryJournals"] = True

        self.review_manager = prescreen_operation.review_manager
        self.settings = self.settings_class(**settings)
        self.local_index = colrev.env.local_index.LocalIndex()

        self.title_complementary_materials_keywords = (
            colrev.env.utils.load_complementary_material_keywords()
        )

    def _conditional_prescreen_entrytypes(
        self, record: colrev.record.record.Record
    ) -> None:
        if self.settings.ENTRYTYPEScope:
            if record.data[Fields.ENTRYTYPE] not in self.settings.ENTRYTYPEScope:
                record.prescreen_exclude(reason="not in ENTRYTYPEScope")

    def _predatory_journal_exclusion(self, record: colrev.record.record.Record) -> None:
        if not self.settings.ExcludePredatoryJournals:
            return
        if Fields.JOURNAL not in record.data:
            return

        rankings = self.local_index.get_journal_rankings(record.data[Fields.JOURNAL])
        if any(x["predatory"] == "yes" for x in rankings):
            record.prescreen_exclude(reason="predatory_journals_beal")

    def _conditional_prescreen_outlets_exclusion(
        self, record: colrev.record.record.Record
    ) -> None:
        if not self.settings.OutletExclusionScope:
            return
        if "values" not in self.settings.OutletExclusionScope:
            return

        for resource in self.settings.OutletExclusionScope["values"]:
            for key, value in resource.items():
                if key in record.data and record.data.get(key, "") == value:
                    record.prescreen_exclude(reason="in OutletExclusionScope")

    def _conditional_prescreen_outlets_inclusion(
        self, record: colrev.record.record.Record
    ) -> None:
        if not self.settings.OutletInclusionScope:
            return

        in_outlet_scope = False
        if "values" in self.settings.OutletInclusionScope:
            for outlet in self.settings.OutletInclusionScope["values"]:
                for key, value in outlet.items():
                    if key in record.data and record.data.get(key, "") == value:
                        in_outlet_scope = True
        if not in_outlet_scope:
            record.prescreen_exclude(reason="not in OutletInclusionScope")

    def _conditional_prescreen_timescope(
        self, record: colrev.record.record.Record
    ) -> None:
        if self.settings.TimeScopeFrom:
            if int(record.data.get(Fields.YEAR, 0)) < self.settings.TimeScopeFrom:
                record.prescreen_exclude(
                    reason="not in TimeScopeFrom " f"(>{self.settings.TimeScopeFrom})"
                )

        if self.settings.TimeScopeTo:
            if int(record.data.get(Fields.YEAR, 5000)) > self.settings.TimeScopeTo:
                record.prescreen_exclude(
                    reason="not in TimeScopeTo " f"(<{self.settings.TimeScopeTo})"
                )

    def _conditional_prescreen_complementary_materials(
        self, record: colrev.record.record.Record
    ) -> None:
        if not self.settings.ExcludeComplementaryMaterials:
            return

        if Fields.TITLE in record.data:
            if (
                record.data[Fields.TITLE].lower()
                in self.title_complementary_materials_keywords
            ):
                record.prescreen_exclude(reason="complementary material")

    def _conditional_presecreen_not_in_ranking(
        self, record: colrev.record.record.Record
    ) -> None:
        if not self.settings.RequireRankedJournals:
            return

        if record.data["journal_ranking"] == "not included in a ranking":
            record.set_status(RecordState.rev_prescreen_excluded)

    def _conditional_prescreen(
        self,
        *,
        record_dict: dict,
    ) -> None:
        if record_dict[Fields.STATUS] != RecordState.md_processed:
            return

        # Note : LanguageScope is covered in prep
        # because dedupe cannot handle merges between languages
        record = colrev.record.record.Record(record_dict)

        self._predatory_journal_exclusion(record=record)
        self._conditional_prescreen_entrytypes(record=record)
        self._conditional_prescreen_outlets_inclusion(record=record)
        self._conditional_prescreen_outlets_exclusion(record=record)
        self._conditional_prescreen_timescope(record=record)
        self._conditional_prescreen_complementary_materials(record=record)
        self._conditional_presecreen_not_in_ranking(record=record)

        if record.data[Fields.STATUS] == RecordState.rev_prescreen_excluded:
            self.review_manager.report_logger.info(
                f" {record.data[Fields.ID]}".ljust(50, " ")
                + "Prescreen excluded (automatically)"
            )
        elif (
            len(self.review_manager.settings.prescreen.prescreen_package_endpoints) == 1
        ):
            record.set_status(RecordState.rev_prescreen_included)
            self.review_manager.report_logger.info(
                f" {record.data[Fields.ID]}".ljust(50, " ")
                + "Prescreen included (automatically)"
            )

    @classmethod
    def add_endpoint(cls, *, operation: colrev.ops.search.Search, params: str) -> None:
        """Add  the scope_prescreen as an endpoint"""

        params_dict = {}
        for p_el in params.split(";"):
            key, value = p_el.split("=")
            params_dict[key] = value

        for (
            existing_scope_prescreen
        ) in operation.review_manager.settings.prescreen.prescreen_package_endpoints:
            if existing_scope_prescreen["endpoint"] != "colrev.scope_prescreen":
                continue
            operation.review_manager.logger.info(
                "Integrating into existing colrev.scope_prescreen"
            )
            for key, value in params_dict.items():
                if (
                    key in existing_scope_prescreen
                    and existing_scope_prescreen[key] != value
                ):
                    operation.review_manager.logger.info(
                        f"Replacing {key} ({existing_scope_prescreen[key]} -> {value})"
                    )
                existing_scope_prescreen[key] = value
            return

        # Insert (if not added before)
        operation.review_manager.settings.prescreen.prescreen_package_endpoints.insert(
            0, {**{"endpoint": "colrev.scope_prescreen"}, **params_dict}
        )

    def run_prescreen(
        self,
        records: dict,
        split: list,  # pylint: disable=unused-argument
    ) -> dict:
        """Prescreen records based on the scope parameters"""

        for record_dict in records.values():
            self._conditional_prescreen(
                record_dict=record_dict,
            )

        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.create_commit(
            msg="Pre-screen (scope)",
            manual_author=False,
        )
        return records



#! /usr/bin/env python
"""Retrieval of PDFs from the unpaywall API"""
from __future__ import annotations

import json
import os
from pathlib import Path

import pymupdf
import requests
import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.packages.unpaywall.src import utils

# pylint: disable=duplicate-code
# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PDFGetInterface)
class Unpaywall:
    """Get PDFs from unpaywall.org"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=False)

    SETTINGS = {
        "email": "packages.pdf_get.colrev.unpaywall.email",
    }

    def __init__(
        self,
        *,
        pdf_get_operation: colrev.ops.pdf_get.PDFGet,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.review_manager = pdf_get_operation.review_manager
        self.pdf_get_operation = pdf_get_operation

        self.email = utils.get_email()

    def _unpaywall(
        self,
        *,
        doi: str,
        retry: int = 0,
        pdfonly: bool = True,
    ) -> str:
        url = f"https://api.unpaywall.org/v2/{doi}"

        try:
            ret = requests.get(url, params={"email": self.email}, timeout=30)
            if ret.status_code == 500 and retry < 3:
                return self._unpaywall(doi=doi, retry=retry + 1)

            if ret.status_code in [404, 500]:
                return "NA"

            best_loc = None
            best_loc = ret.json()["best_oa_location"]

            assert ret.json()["is_oa"]
            assert best_loc is not None
            assert not (pdfonly and best_loc["url_for_pdf"] is None)

        except (
            json.decoder.JSONDecodeError,
            KeyError,
            requests.exceptions.RequestException,
            AssertionError,
        ):
            return "NA"

        return best_loc["url_for_pdf"]

    def _is_pdf(self, *, path_to_file: Path) -> bool:
        with pymupdf.open(path_to_file) as doc:
            doc.load_page(0).get_text()
        return True

    def get_pdf(
        self, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        """Get PDFs from unpaywall"""

        if Fields.DOI not in record.data:
            return record

        pdf_filepath = self.pdf_get_operation.get_target_filepath(record)

        url = self._unpaywall(doi=record.data[Fields.DOI])
        if url == "NA":
            return record
        if "Invalid/unknown DOI" in url:
            return record

        try:
            res = requests.get(
                url,
                headers={
                    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) "
                    "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36"
                },
                stream=True,
                timeout=30,
            )

            if 200 == res.status_code:
                pdf_filepath.parents[0].mkdir(exist_ok=True, parents=True)
                with open(pdf_filepath, "wb") as file:
                    file.write(res.content)
                if self._is_pdf(path_to_file=pdf_filepath):
                    self.review_manager.report_logger.debug(
                        "Retrieved pdf (unpaywall):" f" {pdf_filepath.name}"
                    )
                    source = (
                        f"https://api.unpaywall.org/v2/{record.data['doi']}"
                        + f"?email={self.email}"
                    )
                    record.update_field(
                        key=Fields.FILE, value=str(pdf_filepath), source=source
                    )
                    self.pdf_get_operation.import_pdf(record)

                else:
                    os.remove(pdf_filepath)
            else:
                if Fields.FULLTEXT not in record.data:
                    record.data[Fields.FULLTEXT] = url
                if self.review_manager.verbose_mode:
                    self.review_manager.logger.info(
                        "Unpaywall retrieval error " f"{res.status_code} - {url}"
                    )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.Timeout,
        ):
            pass

        return record



#! /usr/bin/env python
"""Creation of an Obsidian database as part of the data operations"""
from __future__ import annotations

from collections import Counter
from pathlib import Path

import zope.interface
from pydantic import BaseModel
from pydantic import Field

import colrev.env.utils
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields


@zope.interface.implementer(colrev.package_manager.interfaces.DataInterface)
class Obsidian:
    """Export the sample into an Obsidian database"""

    ci_supported: bool = Field(default=False)

    class ObsidianSettings(
        colrev.package_manager.package_settings.DefaultSettings, BaseModel
    ):
        """Settings for Obsidian"""

        endpoint: str
        version: str
        config: dict

        # _details = {
        #     "config": {
        #         "tooltip": "TODO"
        #     },
        # }

    settings_class = ObsidianSettings

    OBSIDIAN_PATH_RELATIVE = Path("data/obsidian")
    OBSIDIAN_PAPER_PATH_RELATIVE = Path("data/obsidian/paper")
    OBSIDIAN_INBOX_PATH_RELATIVE = Path("data/obsidian/inbox.md")
    GITIGNORE_LIST = [
        "data/obsidian/.obsidian/core-plugins.json",
        "data/obsidian/.obsidian/workspace.json",
        "data/obsidian/.obsidian/core-plugins-migration.json",
        "data/obsidian/.obsidian/app.json",
        "data/obsidian/.obsidian/hotkeys.json",
        "data/obsidian/.obsidian/appearance.json",
    ]

    def __init__(
        self,
        *,
        data_operation: colrev.ops.data.Data,
        settings: dict,
    ) -> None:
        self.review_manager = data_operation.review_manager
        self.data_operation = data_operation

        # Set default values (if necessary)
        if "version" not in settings:
            settings["version"] = "0.1"
        if "config" not in settings:
            settings["config"] = {}

        self.settings = self.settings_class(**settings)

        self.endpoint_path = self.review_manager.path / self.OBSIDIAN_PATH_RELATIVE
        self.endpoint_paper_path = (
            self.review_manager.path / self.OBSIDIAN_PAPER_PATH_RELATIVE
        )
        self.endpoint_inbox_path = (
            self.review_manager.path / self.OBSIDIAN_INBOX_PATH_RELATIVE
        )
        if hasattr(self.review_manager, "dataset"):
            self.review_manager.dataset.update_gitignore(add=self.GITIGNORE_LIST)

    # pylint: disable=unused-argument
    @classmethod
    def add_endpoint(cls, operation: colrev.ops.data.Data, params: str) -> None:
        """Add as an endpoint"""

        add_source = {
            "endpoint": "colrev.obsidian",
            "version": "0.1",
            "config": {},
        }

        operation.review_manager.settings.data.data_package_endpoints.append(add_source)

    def _get_obsidian_missing(self, *, included: list) -> list:
        in_obsidian = []
        for md_file in self.endpoint_paper_path.glob("*.md"):
            # missing: if todo in file
            if "#todo" not in md_file.read_text():
                in_obsidian.append(str(md_file.name).replace(".md", ""))
        return [x for x in included if x not in in_obsidian]

    def _get_keywords(self, *, record_dict: dict) -> list:
        keywords = []

        try:
            tei_file = (
                self.review_manager.path
                / colrev.record.record.Record(record_dict).get_tei_filename()
            )

            if tei_file.is_file():
                tei = self.review_manager.get_tei(tei_path=tei_file)
                keywords = [
                    x.lower().replace(" ", "-") for x in tei.get_paper_keywords()
                ]
        except Exception as exc:  # pylint: disable=broad-except
            print(exc)

        if int(record_dict.get(Fields.CITED_BY, 0)) > 100:
            keywords.append("highly_cited")

        return keywords

    def _append_missing_records(self, *, records: dict, silent_mode: bool) -> None:
        included = self.data_operation.get_record_ids_for_synthesis(records)
        missing_records = self._get_obsidian_missing(included=included)
        if len(missing_records) == 0:
            if not silent_mode:
                self.review_manager.logger.info(
                    "All records included. Nothing to export."
                )
            return

        inbox_text = ""
        if self.endpoint_inbox_path.is_file():
            inbox_text = self.endpoint_inbox_path.read_text()
        self.endpoint_inbox_path.parent.mkdir(exist_ok=True)
        if not self.endpoint_inbox_path.is_file():
            with open(self.endpoint_inbox_path, "w", encoding="utf-8") as inbox_file:
                inbox_file.write("Papers to synthesize:\n\n")

        missing_record_entities = {}
        with open(self.endpoint_inbox_path, "a", encoding="utf-8") as inbox_file:
            for missing_record in missing_records:
                if missing_record not in inbox_text:
                    inbox_file.write(f"- [[{missing_record}]]\n")
        for missing_record in missing_records:
            paper_summary_path = self.endpoint_paper_path / Path(f"{missing_record}.md")

            missing_record_entities[paper_summary_path] = {
                Fields.KEYWORDS: self._get_keywords(record_dict=records[missing_record])
            }

        all_keywords = [x[Fields.KEYWORDS] for x in missing_record_entities.values()]
        all_keywords = [item for sublist in all_keywords for item in sublist]

        cnt = Counter(all_keywords)
        frequent_keywords = [k for k, v in cnt.items() if v > 2]

        # Drop research-methods related keywords:
        frequent_keywords = [
            x
            for x in frequent_keywords
            if x
            not in [
                "case-study",
                "design-science",
                "design-science-research",
                "research-agenda",
                "design-theory",
                "mixed-methods",
                "multilevel-analysis",
                "pls",
                "text-minig",
                "associate-editor",
                "chang",
                "ethnography",
                "empirical-research",
                "qualitative-study",
                "ramesh",
                "natural-experiment",
                "simulation",
                "structural-equation-modeling",
                "longitudinal-study",
            ]
        ]

        # for (
        #     paper_summary_path,
        #     missing_record_entity,
        # ) in missing_record_entities.items():
        #     if not paper_summary_path.is_file():
        #         paper_summary_path.parent.mkdir(exist_ok=True, parents=True)
        #         with open(paper_summary_path, "w", encoding="utf-8") as paper_summary:
        #             selected_keywords = [
        #                 x
        #                 for x in missing_record_entity[Fields.KEYWORDS]
        #                 if x in frequent_keywords and x not in ["highly_cited"]
        #             ]
        #             # paper_summary.write(f"#paper {' #'.join(selected_keywords)} #todo\n\n")
        #             paper_summary.write(f"#{' #'.join(selected_keywords)}\n\n")
        #             if "highly_cited" in missing_record_entity[Fields.KEYWORDS]:
        #                 paper_summary.write("highly_cited")

        # later : export to csl-json (based on bibliography_export)
        # (absolute PDF paths, read-only/hidden/gitignored, no provenance fields)

        # self.review_manager.dataset.add_changes(self.OBSIDIAN_INBOX_PATH_RELATIVE)

    def update_data(
        self,
        records: dict,  # pylint: disable=unused-argument
        synthesized_record_status_matrix: dict,  # pylint: disable=unused-argument
        silent_mode: bool,
    ) -> None:
        """Update the obsidian vault"""

        self.review_manager.logger.debug("Export to obsidian endpoint")

        self._append_missing_records(records=records, silent_mode=silent_mode)

    def update_record_status_matrix(
        self,
        synthesized_record_status_matrix: dict,
        endpoint_identifier: str,
    ) -> None:
        """Update the record_status_matrix"""

        missing_records = self._get_obsidian_missing(
            included=list(synthesized_record_status_matrix.keys())
        )

        # Note : automatically set all to True / synthesized
        for syn_id in list(synthesized_record_status_matrix.keys()):
            if syn_id in missing_records:
                synthesized_record_status_matrix[syn_id][endpoint_identifier] = False
            else:
                synthesized_record_status_matrix[syn_id][endpoint_identifier] = True

    def get_advice(
        self,
    ) -> dict:
        """Get advice on the next steps (for display in the colrev status)"""

        data_endpoint = "Data operation [obisdian data endpoint]: "

        advice = {
            "msg": f"{data_endpoint}"
            + "\n    - New records are added to the obsidian vault (data/obsidian)",
            "detailed_msg": "TODO",
        }
        return advice



#! /usr/bin/env python
"""SearchSource: Web of Science"""
from __future__ import annotations

from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class WebOfScienceSearchSource:
    """Web of Science"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    endpoint = "colrev.web_of_science"
    source_identifier = (
        "https://www.webofscience.com/wos/woscc/full-record/" + "{{unique-id}}"
    )
    search_types = [SearchType.DB]

    ci_supported: bool = Field(default=False)
    heuristic_status = SearchSourceHeuristicStatus.supported

    db_url = "http://webofscience.com/"

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.source_operation = source_operation
        self.review_manager = source_operation.review_manager

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Web of Science"""

        result = {"confidence": 0.0}

        if data.count("UT WOS:") > 0.4 * data.count("TI "):
            result["confidence"] = 0.7
            return result

        if "Unique-ID = {WOS:" in data:
            result["confidence"] = 0.7
            return result
        if "UT_(Unique_WOS_ID) = {WOS:" in data:
            result["confidence"] = 0.7
            return result
        if "@article{ WOS:" in data or "@article{WOS:" in data:
            if data.count("{WOS:") > data.count("\n@"):
                result["confidence"] = 1.0
            elif data.count("{ WOS:") > data.count("\n@"):
                result["confidence"] = 1.0
            else:
                result["confidence"] = 0.7

            return result

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        search_source = operation.create_db_source(
            search_source_cls=cls,
            params={},
        )
        operation.add_source_and_search(search_source)
        return search_source

    def search(self, rerun: bool) -> None:
        """Run a search of WebOfScience"""

        if self.search_source.search_type == SearchType.DB:
            self.source_operation.run_db_search(  # type: ignore
                search_source_cls=self.__class__,
                source=self.search_source,
            )
            return

        raise NotImplementedError

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def _load_bib(self) -> dict:
        def field_mapper(record_dict: dict) -> None:
            for key in list(record_dict.keys()):
                if key not in ["ID", "ENTRYTYPE"]:
                    record_dict[key.lower()] = record_dict.pop(key)
            record_dict.pop("book-group-author", None)
            record_dict.pop("organization", None)
            record_dict.pop("researcherid-numbers", None)
            if "eissn" in record_dict:
                if Fields.ISSN not in record_dict:
                    record_dict[Fields.ISSN] = record_dict.pop("eissn", "")
                else:
                    record_dict[Fields.ISSN] += ";" + record_dict.pop("eissn", "")
            if "note" in record_dict:
                record_dict[f"{self.endpoint}.note"] = record_dict.pop("note")
            if "earlyaccessdate" in record_dict:
                record_dict[f"{self.endpoint}.earlyaccessdate"] = record_dict.pop(
                    "earlyaccessdate"
                )
            if "article-number" in record_dict:
                record_dict[f"{self.endpoint}.article-number"] = record_dict.pop(
                    "article-number"
                )
            if "orcid-numbers" in record_dict:
                record_dict[f"{self.endpoint}.orcid-numbers"] = record_dict.pop(
                    "orcid-numbers"
                )
            if "unique-id" in record_dict:
                record_dict[f"{self.endpoint}.unique-id"] = record_dict.pop("unique-id")
            if "book-author" in record_dict:
                record_dict[f"{self.endpoint}.book-author"] = record_dict.pop(
                    "book-author"
                )

        records = colrev.loader.load_utils.load(
            filename=self.search_source.filename,
            logger=self.review_manager.logger,
            unique_id_field="ID",
            field_mapper=field_mapper,
        )
        return records

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            return self._load_bib()

        raise NotImplementedError

    def prepare(
        self,
        record: colrev.record.record_prep.PrepRecord,
        source: colrev.settings.SearchSource,
    ) -> colrev.record.record.Record:
        """Source-specific preparation for Web of Science"""

        # pylint: disable=colrev-missed-constant-usage
        record.format_if_mostly_upper(Fields.TITLE, case="sentence")
        record.format_if_mostly_upper(Fields.JOURNAL, case="title")
        record.format_if_mostly_upper(Fields.BOOKTITLE, case="title")
        record.format_if_mostly_upper(Fields.AUTHOR, case="title")

        record.remove_field(key="colrev.web_of_science.researcherid-numbers")
        record.remove_field(key="colrev.web_of_science.orcid-numbers")
        record.remove_field(key="colrev.web_of_science.book-group-author")
        record.remove_field(key="colrev.web_of_science.note")
        record.remove_field(key="colrev.web_of_science.organization")
        record.remove_field(key="colrev.web_of_science.eissn")
        record.remove_field(key="colrev.web_of_science.earlyaccessdate")

        record.remove_field(key="colrev.web_of_science.meeting")
        record.remove_field(key="colrev.web_of_science.article-number")

        if record.data[Fields.AUTHOR] == "[Anonymous]":
            del record.data[Fields.AUTHOR]
            record.add_field_provenance(
                key=Fields.AUTHOR, source="web_of_scienc.prep", note="IGNORE:missing"
            )

        return record



#! /usr/bin/env python
"""Open Alex API"""
import pyalex
from pyalex import Works

import colrev.env.language_service
import colrev.exceptions as colrev_exceptions
import colrev.record.record_prep
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields
from colrev.constants import FieldValues

# pylint: disable=too-few-public-methods


class OpenAlexAPI:
    """Connector for the Open Alex API"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        email: str,
        # session: requests.Session,
        # query: typing.Optional[str] = None,
        # timeout: int = 60,
    ):
        pyalex.config.email = email
        self.language_service = colrev.env.language_service.LanguageService()

    def _fix_errors(self, *, record: colrev.record.record.Record) -> None:
        if "PubMed" == record.data.get(Fields.JOURNAL, ""):
            record.remove_field(key=Fields.JOURNAL)
        try:
            self.language_service.unify_to_iso_639_3_language_codes(record=record)
        except colrev_exceptions.InvalidLanguageCodeException:
            record.remove_field(key=Fields.LANGUAGE)

    def _set_author_from_item(self, *, record_dict: dict, item: dict) -> None:
        author_list = []
        # pylint: disable=colrev-missed-constant-usage
        for author in item["authorships"]:
            if "author" not in author:
                continue
            if author["author"].get("display_name", None) is None:
                continue
            author_string = colrev.record.record_prep.PrepRecord.format_author_field(
                author["author"]["display_name"]
            )
            author_list.append(author_string)

        record_dict[Fields.AUTHOR] = " and ".join(author_list)

    def _parse_item_to_record(self, *, item: dict) -> colrev.record.record.Record:
        def set_entrytype(*, record_dict: dict, item: dict) -> None:
            # pylint: disable=colrev-missed-constant-usage
            if "title" in record_dict and record_dict["title"] is None:
                del record_dict["title"]
            if item.get("type_crossref", "") == "proceedings-article":
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.INPROCEEDINGS
                if (
                    item.get("primary_location", None) is not None
                    and item["primary_location"].get("source", None) is not None
                ):
                    display_name = item["primary_location"]["source"]["display_name"]
                    if display_name != "Proceedings":
                        record_dict[Fields.BOOKTITLE] = display_name
            elif item["type"] in ["journal-article", "article"]:
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.ARTICLE
                if (
                    item.get("primary_location", None) is not None
                    and item["primary_location"].get("source", None) is not None
                ):
                    record_dict[Fields.JOURNAL] = item["primary_location"]["source"][
                        "display_name"
                    ]
            else:
                record_dict[Fields.ENTRYTYPE] = ENTRYTYPES.MISC

        record_dict = {}
        record_dict["id"] = item["id"].replace("https://openalex.org/", "")
        # pylint: disable=colrev-missed-constant-usage
        if "title" in item and item["title"] is not None:
            record_dict[Fields.TITLE] = item["title"].lstrip("[").rstrip("].")
        set_entrytype(record_dict=record_dict, item=item)

        if "publication_year" in item and item["publication_year"] is not None:
            record_dict[Fields.YEAR] = str(item["publication_year"])
        # pylint: disable=colrev-missed-constant-usage
        if "language" in item and item["language"] is not None:
            record_dict[Fields.LANGUAGE] = item["language"]

        if "is_retracted" in item and item["is_retracted"]:
            record_dict[FieldValues.RETRACTED] = item["is_retracted"]

        # pylint: disable=colrev-missed-constant-usage
        if "doi" in item and item["doi"] is not None:
            record_dict[Fields.DOI] = (
                item["doi"].upper().replace("HTTPS://DOI.ORG/", "")
            )

        record_dict[Fields.CITED_BY] = item["cited_by_count"]

        # pylint: disable=colrev-missed-constant-usage
        if "volume" in item["biblio"] and item["biblio"]["volume"] is not None:
            record_dict[Fields.VOLUME] = item["biblio"]["volume"]
        if "issue" in item["biblio"] and item["biblio"]["issue"] is not None:
            record_dict[Fields.NUMBER] = item["biblio"]["issue"]
        if "first_page" in item["biblio"] and item["biblio"]["first_page"] is not None:
            record_dict[Fields.PAGES] = item["biblio"]["first_page"]
        if "last_page" in item["biblio"] and item["biblio"]["last_page"] is not None:
            record_dict[Fields.PAGES] += "--" + item["biblio"]["last_page"]

        self._set_author_from_item(record_dict=record_dict, item=item)
        record = colrev.record.record.Record(record_dict)

        self._fix_errors(record=record)
        return record

    def get_record(self, *, open_alex_id: str) -> colrev.record.record.Record:
        """Get a record from OpenAlex"""

        item = Works()[open_alex_id]
        retrieved_record = self._parse_item_to_record(item=item)
        return retrieved_record
